#ifndef __ARRAY_H
#define __ARRAY_H

#define OPTIMIZE_FAST_ARRAYS
//#define STDMAP_KEYS

#include "AnsiString.h"
#include "ConceptTypes.h"
#include "AnsiList.h"
#include "GarbageCollector.h"
#include "ConceptPools.h"
#ifdef STDMAP_KEYS
    #include <map>
#endif

#define     KEY_NUMBER    0

#define     ARRAY_INCREMENT    0x20
#define     KEY_INCREMENT      0x20

#define     MAX_DIRTY_DELTA    0x80
#define     DIRTY_TRESHOLD     0x1000

#define ArrayElement           VariableDATA *

#ifdef STDMAP_KEYS
    struct TinyCmp {
       inline bool operator()(const char *a, const char *b) const {
           return strcmp(a, b) < 0;
       }
    };

    typedef std::map<const char *, intptr_t, TinyCmp> KeyMap;
    typedef std::pair<const char *, intptr_t> KeyMapPair;
#else
    typedef struct _KEY {
        char *KEY;
        intptr_t index;
        signed char dirty;

        POOLED(_KEY)
    } ArrayKey;
#endif

typedef struct _NODE {
    ArrayElement ELEMENTS[ARRAY_INCREMENT];
    _NODE        *NEXT;

    POOLED(_NODE)
} NODE;

#ifdef OPTIMIZE_FAST_ARRAYS
    #define FAST_CHECK(ARR, index)  (((ARR)->cached_data) && (index<(int)8192))  ? (ARR)->cached_data[index] : (VariableDATA *)0
#endif

class Array {
    friend int MarkRecursive(void *PIF, Array *arr, char reach_id_flag);
    friend int GetMemoryStatistics(void *PIF, void *RESULT);
private:
#ifdef OPTIMIZE_FAST_ARRAYS
    intptr_t LastNodeIndex;
    NODE *LastNode;
#endif

#ifdef STDMAP_KEYS
    KeyMap *Keys;
#else
    ArrayKey * Keys;
    intptr_t LastKey;
    intptr_t KeysCount;
#endif

    NODE *FIRST;
    NODE *LAST;
    intptr_t NODE_COUNT;
    intptr_t COUNT;

    void *PIF;
    intptr_t AddKey(AnsiString *KEY, intptr_t index);
#ifndef STDMAP_KEYS
    intptr_t FindPlace(char *KEY, intptr_t *in_dirty_zone=NULL);
    void CleanIndex(bool forced=false);
#endif

public:
#ifdef OPTIMIZE_FAST_ARRAYS
    VariableDATA **cached_data;
#endif
    POOLED(Array)

    intptr_t FindKey(char *KEY);
    intptr_t FindIndex(intptr_t index);

    int LINKS;
    intptr_t Count();

    void __GO_GARBAGE(void *PIF, GarbageCollector *__gc_obj, GarbageCollector *__gc_array, GarbageCollector *__gc_vars);

    void UnlinkObjects();

    VariableDATA *NewArray();

    char        *GetKey(intptr_t);
    intptr_t         GetIndex(char *Key);

    VariableDATA *Get(intptr_t);
    VariableDATA *Get(VariableDATA *KEY);
    // for JIT
    VariableDATA *GetWithCreate(intptr_t i, NUMBER default_value = 0);
    VariableDATA *Add(VariableDATA *VAR_TO_ADD);
    VariableDATA *AddCopy(VariableDATA *VAR_TO_ADD);

    VariableDATA *ModuleGet(char *key);
    VariableDATA *ModuleGet(intptr_t);

    Array *SortArrayElementsByKey();

    AnsiString ToString(int level = 0, Array *parent = 0, int parents = 0);
    Array *Concat(Array *array);

    void EnsureSize(intptr_t size, VariableDATA *default_value);
    Array(void *PIF);
    ~Array();

    char flags;
    char reachable;
    inline void operator delete (void *obj) {
        FreeArray(obj);
    }
};

typedef struct {
	Array POOL[ARRAY_POOL_BLOCK_SIZE];
	void *NEXT;
	void *PREV;
    void *PIF;
	unsigned char POOL_VARS;
    unsigned char FIRST_VAR;
} ArrayPool;

#endif // __ARRAY_H
