#include "CompiledClass.h"
#include "ClassMember.h"
#include "ConceptInterpreter.h"


POOLED_IMPLEMENTATION(CompiledClass)

CompiledClass::CompiledClass(ClassCode *CC) {
    _Class   = CC;
    _CONTEXT = 0;
    LINKS    = 0;
    //---------------------------------
    INTEGER Count  = _Class->DataMembersCount;
    if (Count) {
        _CONTEXT = (VariableDATA **)FAST_MALLOC(sizeof(VariableDATA*) * Count);
        memset(_CONTEXT, 0, sizeof(VariableDATA*) * Count);
        //for (register INTEGER i = 0; i < Count; i++)
        //    _CONTEXT[i] = 0;
    }
}

VariableDATA *CompiledClass::CreateVariable(INTEGER reloc, ClassMember *CM) {
    VariableDATA *_CONTEXT_i=_CONTEXT[reloc];
    if (!_CONTEXT_i) {
        if ((CM) && (!CM->IS_FUNCTION)) {
            _CONTEXT_i                     = (VariableDATA *)VAR_ALLOC(_Class->PIF);
            _CONTEXT[reloc] = _CONTEXT_i;
            if (CM->VD) {
                if (CM->VD->TYPE == VARIABLE_ARRAY) {
                    _CONTEXT_i->CLASS_DATA = new(AllocArray(_Class->PIF)) Array(_Class->PIF);
                } else
                if (CM->VD->TYPE == VARIABLE_STRING) {
                    _CONTEXT_i->CLASS_DATA = NULL;
                    CONCEPT_STRING(_CONTEXT_i) = CM->VD->value;
                } else
                    _CONTEXT_i->NUMBER_DATA = CM->VD->nValue;

                _CONTEXT_i->LINKS       = 1;
                _CONTEXT_i->TYPE        = CM->VD->TYPE;

                _CONTEXT_i->IS_PROPERTY_RESULT = 0;
            } else {
                _CONTEXT_i->LINKS              = 1;
                _CONTEXT_i->NUMBER_DATA        = 0;
                _CONTEXT_i->TYPE               = VARIABLE_NUMBER;
                _CONTEXT_i->IS_PROPERTY_RESULT = 0;
            }
        }
    }
    return _CONTEXT_i;
}

AnsiString CompiledClass::GetClassName() {
    if (_Class) {
        return(_Class->NAME);
    }
    return(NULL_STRING);
}

void CompiledClass::__GO_GARBAGE(void *PIF, GarbageCollector *__gc_obj, GarbageCollector *__gc_array, GarbageCollector *__gc_vars) {
    if (this->LINKS<0)
        return;

    // ensure is not deleted
    this->LINKS++;
    if (_Class->DESTRUCTOR)
        Destroy((PIFAlizator *)PIF);

    if (!this->_CONTEXT)
        return;

    VariableDATA **r_CONTEXT = this->_CONTEXT;
    this->_CONTEXT = NULL;

    INTEGER Count = _Class->DataMembersCount;

    for (INTEGER i = 0; i < Count; i++) {
        if (r_CONTEXT [i]) {
            // ensure is not deleted
            r_CONTEXT [i]->LINKS++;
            __gc_vars->Reference(r_CONTEXT[i]);
            void *orig_data=r_CONTEXT [i]->CLASS_DATA;
            if (orig_data) {
                if ((r_CONTEXT [i]->TYPE == VARIABLE_CLASS) || (r_CONTEXT [i]->TYPE == VARIABLE_DELEGATE)) {
                    if (orig_data!=this) {
                        __gc_obj->Reference(orig_data);
                        ((CompiledClass *)orig_data)->__GO_GARBAGE(PIF, __gc_obj, __gc_array, __gc_vars);
                    }
                    orig_data=0;
                } else
                if (r_CONTEXT [i]->TYPE == VARIABLE_ARRAY) {
                    __gc_array->Reference(orig_data);
                    ((Array *)orig_data)->__GO_GARBAGE(PIF, __gc_obj, __gc_array, __gc_vars);
                    orig_data=0;
                }
            }
        }
    }
    FAST_FREE(r_CONTEXT);
    this->LINKS = -1;
}

ClassCode *CompiledClass::GetClass() {
    return(_Class);
}

VariableDATA **CompiledClass::GetContext() {
    return(_CONTEXT);
}

int CompiledClass::Destroy(PIFAlizator *PIF) {
    VariableDATA *OWNER = (VariableDATA *)VAR_ALLOC(PIF);

    OWNER->TYPE       = VARIABLE_CLASS;
    OWNER->LINKS      = 1;
    OWNER->CLASS_DATA = this;
    //OWNER->ACCESS=ACCESS_LOCAL;
    OWNER->IS_PROPERTY_RESULT = 0;
    this->LINKS++;
    VariableDATA *THROW_DATA = 0;

    STACK(NULL, _Class->DESTRUCTOR_MEMBER->_DEBUG_STARTLINE)
    VariableDATA * RESULT = _Class->DESTRUCTOR_MEMBER->Execute(PIF, this->_Class->CLSID, OWNER, 0, _CONTEXT, THROW_DATA, NULL);
    UNSTACK;
    if (RESULT) {
        FREE_VARIABLE(RESULT);
    }

    // avoid double deleting of the class !!!
    OWNER->TYPE       = VARIABLE_NUMBER;
    OWNER->CLASS_DATA = 0;
    FREE_VARIABLE(OWNER);

    if (THROW_DATA) {
        FREE_VARIABLE(THROW_DATA);
        AnsiException *Exc = new AnsiException(ERR635, 0, 635, 0, _Class->_DEBUG_INFO_FILENAME, _Class->NAME, _Class->DESTRUCTOR_MEMBER->NAME);
        _Class->PIF->AcknoledgeRunTimeError(NULL, Exc);
    }
    this->LINKS = -1;
    return(1);
}

int CompiledClass::HasDestructor() {
    if (this->LINKS < 0) {
        return(0);
    }

    if (_Class->DESTRUCTOR) {
        return(1);
    }

    return(0);
}

void CompiledClass::UnlinkObjects() {
    if (this->LINKS < 0) {                                                                                                                                                                                                                                                                         //(!_CONTEXT)
        return;
    }
    this->LINKS = -1;

    /*if (_Class->DESTRUCTOR)
     *  Destroy();*/

    INTEGER Count = _Class->DataMembersCount;     //_Class->pMEMBERS_COUNT;

    for (register INTEGER i = 0; i < Count; i++) {
        VariableDATA *_CONTEXT_i = _CONTEXT [i];
        if (_CONTEXT_i) {
            // delete no object ! ... it's the garbage collector's job
            if (_CONTEXT_i->CLASS_DATA) {
                //_CONTEXT_i->TYPE       = VARIABLE_NUMBER;
                //_CONTEXT_i->CLASS_DATA = 0;

                //FREE_VARIABLE(_CONTEXT_i);
                FREE_VARIABLE_NO_OBJECTS(_CONTEXT_i);
            }
        }
    }

    FAST_FREE(_CONTEXT);
    //delete[] _CONTEXT;
    _CONTEXT = 0;
}

CompiledClass::~CompiledClass() {
    //this->_Class->PIF->__gc_obj.Unreference(this);
    if (this->LINKS < 0)
        return;

    if (_Class->DESTRUCTOR) {
#ifdef SIMPLE_MULTI_THREADING
        PIFAlizator *PIF=this->_Class->PIF;
        char master_lock=PIF->MasterLock;
        char IsWriteLocked=master_lock;
        if (master_lock) {
            WRITE_UNLOCK
        }
#endif

        Destroy(/*this->PIF*/_Class->PIF);

#ifdef SIMPLE_MULTI_THREADING
        if (master_lock) {
            WRITE_LOCK
        }
#endif
    }

    this->LINKS = -1;
    INTEGER Count = _Class->DataMembersCount;

    for (register INTEGER i = 0; i < Count; i++) {
        VariableDATA *_CONTEXT_i = _CONTEXT [i];
        if (_CONTEXT_i) {
            //=======================================//
            if ((_CONTEXT_i->CLASS_DATA == this) && ((_CONTEXT_i->TYPE == VARIABLE_CLASS) || (_CONTEXT_i->TYPE == VARIABLE_DELEGATE))) {
                _CONTEXT_i->TYPE       = VARIABLE_NUMBER;
                _CONTEXT_i->CLASS_DATA = 0;
            }
            //=======================================//
            FREE_VARIABLE(_CONTEXT_i);
        }
    }
    FAST_FREE(_CONTEXT);
}
