#include "Codes.h"
#include "ConceptTypes.h"
#include "AnsiString.h"
#include "CompiledClass.h"
#include "Debugger.h"
#include "ConceptInterpreter.h"
#include <string.h>
#include "ModuleLink.h"

#define MIN(a, b)    (((a) < (b)) ? (a) : (b))

VariableDATA *VarToClean [0xFF];
int          VarCnt = 0;

#define CLEAN_VARS    { for (int i = 0; i < VarCnt; i ++) { FREE_VARIABLE(VarToClean [i]); } VarCnt = 0; }
#define ADD_VAR(VAR)    { if (VAR) VarToClean [VarCnt ++] = VAR; }

VariableDATA *GetClassMember(void *CLASS_PTR, char *class_member_name) {
    if (CLASS_PTR) {
        ClassCode    *CCode    = ((CompiledClass *)CLASS_PTR)->GetClass();
        VariableDATA **CONTEXT = ((CompiledClass *)CLASS_PTR)->GetContext();
        if ((CCode) && (CONTEXT)) {
            INTEGER index = CCode->HasMemberInCompiled(class_member_name);
            if (index) {
                index--;
                ClassMember *CM = index < CCode->pMEMBERS_COUNT ? CCode->pMEMBERS[index] : 0;

                if ((CM) && (CM->IS_FUNCTION)) {
                    //VariableDATA
                    index = CCode->GetAbsoluteMemberID(index + 1) - 1;

                    ParamList PLIST;
                    PLIST.COUNT = 0;

                    AnalizerElement AE;
                    AE._DEBUG_INFO_LINE = AE.ID = AE.TYPE = 0;

                    VariableDATA *Owner = (VariableDATA *)VAR_ALLOC(/*((CompiledClass *)CLASS_PTR)->PIF*/CCode->PIF);
                    //Owner->CALL_SET           = 0;
                    Owner->CLASS_DATA         = CLASS_PTR;
                    Owner->IS_PROPERTY_RESULT = 0;
                    Owner->LINKS       = 1;
                    //Owner->NUMBER_DATA = 0;
                    Owner->TYPE        = VARIABLE_CLASS;
                    ((CompiledClass *)CLASS_PTR)->LINKS++;

                    try {
                        VariableDATA *VarDATA = CCode->ExecuteMember(/*((CompiledClass *)CLASS_PTR)->PIF*/CCode->PIF,
                            index,
                            Owner,
                            &AE,
                            true,
                            CM->IS_FUNCTION == 3 ? 0 : &PLIST,
                            (VariableDATA **)-1,
                            0,
                            CCode->CLSID,
                            CCode->CLSID,
                            NULL
                            );
                        FREE_VARIABLE(Owner);
                        ADD_VAR(VarDATA);
                        return(VarDATA);
                    } catch (VariableDATA *THROW_DATA) {
                        FREE_VARIABLE(THROW_DATA);
                        FREE_VARIABLE(Owner);
                        return(0);
                    }
                } else {
                    if (CCode->RELOCATIONS2) {
                        index = CCode->RELOCATIONS2 [index] - 1;
                        if (index>=0) {
                            if ((CONTEXT) && (!CONTEXT [index])) {
                                if ((CM->VD) && ((CM->VD->TYPE!=VARIABLE_NUMBER) || (CM->VD->nValue)))
                                    ((CompiledClass *)CLASS_PTR)->CreateVariable(index, CM);
                            }
                            return(CONTEXT [index]);
                        }
                    }
                }
            }
        }
    }
    return(0);
}

VariableDATA *GetClassMembers(void *CLASS_PTR, char *members_string) {
    AnsiString ParsedVariableName = members_string;
    AnsiString _S(".");
    int        point            = ParsedVariableName.Pos(_S);
    void       *CLASS_PTR_LOCAL = CLASS_PTR;

    while (point > 0) {
        AnsiString temp_varname = ParsedVariableName;
        AnsiString temp2        = temp_varname;
        temp2.c_str() [point - 1] = 0;
        temp_varname       = (char *)temp2.c_str();
        temp2              = ParsedVariableName.c_str() + point;
        ParsedVariableName = temp2;
        VariableDATA *Temp = GetClassMember(CLASS_PTR_LOCAL, temp_varname);
        if ((Temp) && (Temp->TYPE == VARIABLE_CLASS)) {
            CLASS_PTR_LOCAL = Temp->CLASS_DATA;
        } else {
            return(0);
        }
        if (!CLASS_PTR_LOCAL) {
            return(0);
        }
        point = ParsedVariableName.Pos(_S);
    }
    return(CLASS_PTR_LOCAL ? GetClassMember(CLASS_PTR_LOCAL, ParsedVariableName) : 0);
}

int GetVariableByName(int operation, void **VDESC, void **CONTEXT, int Depth, char *VariableName, char *buffer, int buf_size, void *PIF, void *STACK_TRACE) {
    if (operation == -1) {
        AnsiString cstack;
        INTERNAL_LOCK(((PIFAlizator *)PIF))
        int first = 1;
        while (STACK_TRACE) {
            ClassMember *CM = (ClassMember *)((SCStack *)STACK_TRACE)->CM;
            if (CM) {
                if (CM->Defined_In) {
                    cstack += ((ClassCode *)CM->Defined_In)->NAME.c_str();
                    cstack += ".";
                } else
                    cstack += "::";
                cstack += CM->NAME;
            } else {
                cstack += "STATIC/LIBRARY.STATIC_FUNCTION";
            }
            cstack += ":";
            if (first) {
                cstack += AnsiString((D_LONG_TYPE)CM->_DEBUG_STARTLINE);
                first = 0;
            } else
                cstack += AnsiString((D_LONG_TYPE)((SCStack *)STACK_TRACE)->line);
            cstack += "\n";
            STACK_TRACE = (SCStack *)((SCStack *)STACK_TRACE)->PREV;
        }
        INTERNAL_UNLOCK(((PIFAlizator *)PIF))
        buffer [buf_size - 1] = 0;
        strncpy(buffer, cstack.c_str(), buf_size - 1);
        buffer [cstack.Length()] = 0;
        return(-1);
    }

    if (!CONTEXT) {
        return(0);
    }
    CLEAN_VARS;

    AnsiString ParsedVariableName = VariableName;
    AnsiString temp_varname       = VariableName;
    int        point = ParsedVariableName.Pos(AnsiString("."));
    if (point > 0) {
        temp_varname = ParsedVariableName;
        AnsiString temp2 = temp_varname;
        temp2.c_str() [point - 1] = 0;
        temp_varname       = (char *)temp2.c_str();
        temp2              = ParsedVariableName.c_str() + point;
        ParsedVariableName = temp2;
    } else {
        point = 0;
    }

    for (INTEGER i = 0; i < Depth; i++) {
        if ((((VariableDESCRIPTOR *)VDESC) [i]).name == temp_varname) {
            if (((VariableDATA *)CONTEXT [i])->TYPE == VARIABLE_STRING) {
                if (operation == 0) {
                    strncpy(buffer, CONCEPT_STRING(((VariableDATA *)CONTEXT [i])).c_str(), buf_size);
                } else
                if (operation == 1) {
                    CONCEPT_STRING(((VariableDATA *)CONTEXT [i])) = buffer;
                }

                CLEAN_VARS;

                return(point ? 0 : VARIABLE_STRING);
            } else
            if (((VariableDATA *)CONTEXT [i])->TYPE == VARIABLE_NUMBER) {
                if (operation == 0) {
                    strncpy(buffer, AnsiString(((VariableDATA *)CONTEXT [i])->NUMBER_DATA).c_str(), buf_size);
                } else
                if (operation == 1) {
                    ((VariableDATA *)CONTEXT [i])->NUMBER_DATA = AnsiString(buffer).ToFloat();
                }

                CLEAN_VARS;

                return(point ? 0 : VARIABLE_NUMBER);
            } else
            if (((VariableDATA *)CONTEXT [i])->TYPE == VARIABLE_CLASS) {
                if (operation == 0) {
                    strncpy(buffer, ((CompiledClass *)((VariableDATA *)CONTEXT [i])->CLASS_DATA)->GetClassName().c_str(), buf_size);
                }
                if (!point) {
                    CLEAN_VARS;
                    return(VARIABLE_CLASS);
                } else {
                    VariableDATA *VD = GetClassMembers(((VariableDATA *)CONTEXT [i])->CLASS_DATA, ParsedVariableName.c_str());
                    if (!VD) {
                        CLEAN_VARS;
                        return(0);
                    }
                    switch (VD->TYPE) {
                        case VARIABLE_STRING:
                            if (operation == 0) {
                                strncpy(buffer, CONCEPT_STRING(VD).c_str(), buf_size);
                            } else
                            if (operation == 1) {
                                CONCEPT_STRING(VD) = buffer;
                                if (VD->IS_PROPERTY_RESULT) {
                                    SetClassMember(((VariableDATA *)CONTEXT [i])->CLASS_DATA, ParsedVariableName.c_str(), VARIABLE_STRING, buffer, 0);
                                }
                            }
                            CLEAN_VARS;
                            return(VD->TYPE);
                            break;

                        case VARIABLE_NUMBER:
                            if (operation == 0) {
                                strncpy(buffer, AnsiString(VD->NUMBER_DATA).c_str(), buf_size);
                            } else
                            if (operation == 1) {
                                VD->NUMBER_DATA = AnsiString(buffer).ToFloat();
                                if (VD->IS_PROPERTY_RESULT) {
                                    SetClassMember(((VariableDATA *)CONTEXT [i])->CLASS_DATA, ParsedVariableName.c_str(), VARIABLE_NUMBER, (char *)"", VD->NUMBER_DATA);
                                }
                            }
                            CLEAN_VARS;
                            return(VD->TYPE);
                            break;

                        case VARIABLE_CLASS:
                            if (operation == 0) {
                                strncpy(buffer, ((CompiledClass *)VD->CLASS_DATA)->GetClassName().c_str(), buf_size);
                            }
                            CLEAN_VARS;
                            return(VD->TYPE);
                            break;

                        case VARIABLE_ARRAY:
                            if (operation == 0) {
                                strncpy(buffer, ((Array *)VD->CLASS_DATA)->ToString(), buf_size);
                            }
                            CLEAN_VARS;
                            return(VD->TYPE);
                            break;

                        case VARIABLE_DELEGATE:
                            {
                                if (operation == 0) {
                                    AnsiString delegate = ((CompiledClass *)VD->CLASS_DATA)->GetClassName();
                                    delegate += (char *)".";
                                    delegate += ((CompiledClass *)VD->CLASS_DATA)->_Class->pMEMBERS [(INTEGER)VD->DELEGATE_DATA]->NAME;
                                    strncpy(buffer, delegate.c_str(), buf_size);
                                }
                                CLEAN_VARS;
                                return(VD->TYPE);
                            }
                            break;

                        default:
                            CLEAN_VARS;
                            return(0);
                    }
                }
            }
            CLEAN_VARS;
            return(0);
        }
    }
    CLEAN_VARS;
    return(0);
}
