#define CACHE_MEMBERS

#include "ModuleLink.h"
#include "AnsiException.h"
#include "Array.h"
#include "ConceptInterpreter.h"
#include "library.h"
#include <string.h>
#include "Codes.h"
#include "PIFAlizator.h"

#ifdef _WIN32
    #include <windows.h>    // LoadLibrary
    #include <io.h>
#else
    #include <unistd.h>
    #include <stdarg.h>
    #include <dlfcn.h>      // dlopen
    #include <poll.h>

    int eof(int fd) {
        struct pollfd ufds[1];
        ufds[0].fd = fd;
        ufds[0].events = POLLIN;

        int res = poll(ufds, 1, 0);

        return !res;
    }

    ssize_t sock_fd_read(int sock, int *fd) {
        ssize_t size;
        char    buf[16];

        if (fd) {
            *fd=-1;
            struct msghdr   msg;
            struct iovec    iov;
            union {
                struct cmsghdr  cmsghdr;
                char        control[CMSG_SPACE(sizeof (int))];
            } cmsgu;
            struct cmsghdr  *cmsg;

            iov.iov_base = buf;
            iov.iov_len = sizeof(buf);

            msg.msg_name = NULL;
            msg.msg_namelen = 0;
            msg.msg_iov = &iov;
            msg.msg_iovlen = 1;
            msg.msg_control = cmsgu.control;
            msg.msg_controllen = sizeof(cmsgu.control);
            size = recvmsg (sock, &msg, 0);
            if (size < 0)
                return size;

            cmsg = CMSG_FIRSTHDR(&msg);
            if (cmsg && cmsg->cmsg_len == CMSG_LEN(sizeof(int))) {
                if (cmsg->cmsg_level != SOL_SOCKET)
                    return -2;

                if (cmsg->cmsg_type != SCM_RIGHTS)
                    return -3;

                *fd = *((int *) CMSG_DATA(cmsg));
            } else
                *fd = -1;
        } else {
            size = read (sock, buf, sizeof(buf));
            if (size < 0)
                return size;
        }
        return size;
    }
#endif

#ifdef CACHE_MEMBERS
#include <map>
static std::map<unsigned int, SYS_INT> FunctionMap;
#endif

INTEGER ImportModule(AnsiString& MODULE_MASK, AnsiList *Errors, INTEGER line, AnsiString FILENAME, AnsiList *TARGET, PIFAlizator *Sender, INTEGER _NO_ERROR_REPORT) {
#ifdef _WIN32
    // windows module
    HANDLE hLIBRARY = LoadLibrary(MODULE_MASK + MODULE_EXTENSION);
#else
    // linux module
    HANDLE hLIBRARY = dlopen(MODULE_MASK + MODULE_EXTENSION, RTLD_LAZY);
#endif
    if (!hLIBRARY) {
        if (!_NO_ERROR_REPORT)
#ifdef _WIN32
        {
            Errors->Add(new AnsiException(ERR670, line, 670, MODULE_MASK, FILENAME), DATA_EXCEPTION);
        }

#else
        {
            Errors->Add(new AnsiException(AnsiString(ERR670) + AnsiString(": ") + (char *)dlerror(), line, 670, MODULE_MASK, FILENAME), DATA_EXCEPTION);
        }
#endif

        else {
            return(0);
        }
    } else {
        TARGET->Add(hLIBRARY, DATA_HANDLER);
        // --------------- Module Management functions -----------------------------------------//
        AnsiString PREFIX = (char *)"_";
#ifdef _WIN32
        CONCEPT_MANAGEMENT_CALL _PROC_ADR = (CONCEPT_MANAGEMENT_CALL)GetProcAddress((HMODULE)hLIBRARY, ON_CREATE);
#else
        CONCEPT_MANAGEMENT_CALL _PROC_ADR = (CONCEPT_MANAGEMENT_CALL)dlsym((HMODULE)hLIBRARY, ON_CREATE);
#endif

        if (!_PROC_ADR) {
#ifdef _WIN32
            _PROC_ADR = (CONCEPT_MANAGEMENT_CALL)GetProcAddress((HMODULE)hLIBRARY, PREFIX + ON_CREATE);
#else
            _PROC_ADR = (CONCEPT_MANAGEMENT_CALL)dlsym((HMODULE)hLIBRARY, PREFIX + ON_CREATE);
#endif
        }
        if (_PROC_ADR) {
            char *LOAD_RESULT = (char *)_PROC_ADR((SYS_INT)hLIBRARY, Sender, Invoke);
            if (LOAD_RESULT) {
                Errors->Add(new AnsiException(ERR880, line, 880, LOAD_RESULT, FILENAME), DATA_EXCEPTION);
            }
        }
        // -------------------------------------------------------------------------------------//
    }
    return(1);
}

INTEGER UnImportModule(HMODULE hMODULE, PIFAlizator *Sender) {
    // --------------- Module Management functions -----------------------------------------//
    AnsiString PREFIX = (char *)"_";

#ifdef _WIN32
    CONCEPT_MANAGEMENT_CALL _PROC_ADR = (CONCEPT_MANAGEMENT_CALL)GetProcAddress((HMODULE)hMODULE, ON_DESTROY);
#else
    CONCEPT_MANAGEMENT_CALL _PROC_ADR = (CONCEPT_MANAGEMENT_CALL)dlsym((HMODULE)hMODULE, ON_DESTROY);
#endif


    PREFIX += ON_DESTROY;
    if (!_PROC_ADR) {
#ifdef _WIN32
        _PROC_ADR = (CONCEPT_MANAGEMENT_CALL)GetProcAddress((HMODULE)hMODULE, PREFIX.c_str() /*+ON_DESTROY*/);
#else
        _PROC_ADR = (CONCEPT_MANAGEMENT_CALL)dlsym((HMODULE)hMODULE, PREFIX.c_str() /*+ON_DESTROY*/);
#endif
    }

    char *LOAD_RESULT = 0;
    if (_PROC_ADR) {
        LOAD_RESULT = (char *)_PROC_ADR((SYS_INT)hMODULE, Sender, Invoke);
    }
    // TO DO : eventual adauga eroarea de la unload in lista de erori ...
    // -------------------------------------------------------------------------------------//

#ifdef _WIN32
    // windows module
    if (Sender)
        return(0);
    else
        return(FreeLibrary((HMODULE)hMODULE));
#else
    // linux module
    if ((Sender) || (LOAD_RESULT))
        return(0);
    else
        return(dlclose((HMODULE)hMODULE));
#endif
    return(0);
}

INTEGER SetVariable(VariableDATA *VD, INTEGER TYPE, char *STRING_VALUE, NUMBER NUMBER_VALUE) {
    if (!VD)
        return -1;
    // is constant ?
    if (VD->IS_PROPERTY_RESULT == -1)
        return -2;
    //---------------//
    CLASS_CHECK(VD);
    VD->CLASS_DATA = NULL;
    //---------------//
    char copy_buffer = 0;
    if (TYPE == _LINK__STRING_DATA) {
        TYPE        = VARIABLE_STRING;
        copy_buffer = 1;
        if (!(INTEGER)NUMBER_VALUE) {
            STRING_VALUE = 0;
        }
    }
    VD->TYPE = TYPE;
    if (VD->TYPE == VARIABLE_STRING) {
        if (!NUMBER_VALUE) {
            CONCEPT_STRING(VD) = STRING_VALUE;
        } else {
            if (copy_buffer) {
                CONCEPT_STRING(VD).LinkBuffer(STRING_VALUE, (INTEGER)NUMBER_VALUE);
            } else {
                CONCEPT_STRING(VD).LoadBuffer(STRING_VALUE, (INTEGER)NUMBER_VALUE);
            }
        }
    } else
    if ((VD->TYPE == VARIABLE_CLASS) || (VD->TYPE == VARIABLE_ARRAY) || (VD->TYPE == VARIABLE_DELEGATE)) {
        VD->CLASS_DATA  = (void *)STRING_VALUE;
        if ((TYPE == VARIABLE_CLASS) || (TYPE == VARIABLE_DELEGATE)) {
            ((CompiledClass *)VD->CLASS_DATA)->LINKS++;
            if (TYPE == VARIABLE_DELEGATE)
                VD->DELEGATE_DATA = (int)NUMBER_VALUE;
        } else {
            ((Array *)VD->CLASS_DATA)->LINKS++;
        }
    } else {
        VD->NUMBER_DATA = NUMBER_VALUE;
    }
    return(0);
}

INTEGER GetVariable(VariableDATA *VD, INTEGER *TYPE, char **STRING_VALUE, NUMBER *NUMBER_VALUE) {
    if (!VD)
        return -1;
    *TYPE         = VD->TYPE;
    *STRING_VALUE=0;

    if (VD->TYPE == VARIABLE_STRING) {
        *STRING_VALUE = CONCEPT_STRING(VD).c_str();
        *NUMBER_VALUE = CONCEPT_STRING(VD).Length();
    } else
    if ((VD->TYPE == VARIABLE_CLASS) || (VD->TYPE == VARIABLE_ARRAY)) {
        *STRING_VALUE = (char *)VD->CLASS_DATA;
        *NUMBER_VALUE = 0;
    } else
    if (VD->TYPE == VARIABLE_DELEGATE) {
        *STRING_VALUE = (char *)VD->CLASS_DATA;
        *NUMBER_VALUE = VD->DELEGATE_DATA;
    } else {
        *NUMBER_VALUE = VD->NUMBER_DATA;
    }
    return(0);
}

INTEGER GetClassMember(void *CLASS_PTR, char *class_member_name, INTEGER *TYPE, char **STRING_VALUE, NUMBER *NUMBER_VALUE) {
    if (CLASS_PTR) {
        ClassCode    *CCode    = ((CompiledClass *)CLASS_PTR)->GetClass();
        VariableDATA **CONTEXT = ((CompiledClass *)CLASS_PTR)->GetContext();
        //if ((CCode) && (CONTEXT)) {
        if (CCode) {
            INTEGER m_type = 0;
            INTEGER index  = CCode->HasMemberInCompiled(class_member_name, &m_type);
            if (index) {
                index--;

                ClassMember *CM = index < CCode->pMEMBERS_COUNT ? CCode->pMEMBERS[index] : 0;

                if ((CM) && (CM->IS_FUNCTION)) {
                    index = CCode->GetAbsoluteMemberID(index + 1) - 1;
                    ParamList PLIST;
                    PLIST.COUNT = 0;

                    AnalizerElement AE;
                    AE._DEBUG_INFO_LINE = AE.ID = AE.TYPE = 0;

                    VariableDATA *Owner = (VariableDATA *)VAR_ALLOC(/*((CompiledClass *)CLASS_PTR)->PIF*/CCode->PIF);
                    //Owner->CALL_SET           = 0;
                    Owner->CLASS_DATA         = CLASS_PTR;
                    Owner->IS_PROPERTY_RESULT = 0;
                    Owner->LINKS       = 1;
                    Owner->TYPE        = VARIABLE_CLASS;
                    ((CompiledClass *)CLASS_PTR)->LINKS++;

                    try {
                        VariableDATA *VarDATA = CCode->ExecuteMember(/*((CompiledClass *)CLASS_PTR)->PIF*/CCode->PIF,
                            index,
                            Owner,
                            &AE,
                            true,
                            CM->IS_FUNCTION == 3 ? 0 : &PLIST,
                            (VariableDATA **)-1,
                            0,
                            CCode->CLSID,
                            CCode->CLSID,
                            NULL
                            );
                        int result=-1;
                        if (VarDATA) {
                            result = GetVariable(VarDATA, TYPE, STRING_VALUE, NUMBER_VALUE);
                            FREE_VARIABLE(VarDATA);
                        }
                        FREE_VARIABLE(Owner);
                        return(result);
                    } catch (VariableDATA *THROW_DATA) {
                        FREE_VARIABLE(THROW_DATA);
                        FREE_VARIABLE(Owner);
                        return(-1);
                    }
                }
                index = CCode->RELOCATIONS2 [index] - 1;
                if (index >= 0) {
                    if ((CONTEXT) && (!m_type) && (!CONTEXT [index]) && (CM))
                        ((CompiledClass *)CLASS_PTR)->CreateVariable(index, CM);

                    if (CONTEXT [index]) {
                        return(GetVariable(CONTEXT [index], TYPE, STRING_VALUE, NUMBER_VALUE));
                    }
                } else {
                    return(-1);
                }
            }
        }
    }
    return(-1);
}

INTEGER GetClassMemberVariable(void *CLASS_PTR, char *class_member_name, void **ptr) {
    if (CLASS_PTR) {
        ClassCode    *CCode    = ((CompiledClass *)CLASS_PTR)->GetClass();
        VariableDATA **CONTEXT = ((CompiledClass *)CLASS_PTR)->GetContext();
        if ((CCode) && (CONTEXT)) {
            INTEGER m_type = 0;
            INTEGER index  = CCode->HasMemberInCompiled(class_member_name, &m_type);
            if (index) {
                index--;
                ClassMember *CM = index < CCode->pMEMBERS_COUNT ? CCode->pMEMBERS[index] : 0;
                index = CCode->RELOCATIONS2 [index] - 1;
                if ((CONTEXT) && (!m_type) && (!CONTEXT [index]) && (CM))
                    ((CompiledClass *)CLASS_PTR)->CreateVariable(index, CM);

                if (CONTEXT [index]) {
                    *ptr = CONTEXT [index];
                }
            }
            return(0);
        }
    }
    return(-1);
}

INTEGER SetClassMember(void *CLASS_PTR, char *class_member_name, INTEGER TYPE, char *STRING_VALUE, NUMBER NUMBER_VALUE) {
    if (CLASS_PTR) {
        ClassCode    *CCode    = ((CompiledClass *)CLASS_PTR)->GetClass();
        VariableDATA **CONTEXT = ((CompiledClass *)CLASS_PTR)->GetContext();
        if (CCode) {
            INTEGER m_type = 0;
            INTEGER index  = CCode->HasMemberInCompiled(class_member_name, &m_type);
            ClassMember *CM = 0;
            if (index) {
                index--;
                CM = index < CCode->pMEMBERS_COUNT ? CCode->pMEMBERS[index] : 0;

                if ((CM) && (CM->IS_FUNCTION == 3) && (CM->MEMBER_SET)) {
                    index = CCode->GetAbsoluteMemberID(index + 1) - 1;

                    AnalizerElement AE;
                    AE._DEBUG_INFO_LINE = AE.ID = AE.TYPE = 0;

                    VariableDATA *Parameter = (VariableDATA *)VAR_ALLOC(/*((CompiledClass *)CLASS_PTR)->PIF*/CCode->PIF);
                    //Parameter->CALL_SET           = 0;
                    Parameter->IS_PROPERTY_RESULT = 0;
                    Parameter->LINKS       = 1;
                    Parameter->TYPE        = TYPE;

                    VariableDATA *CONTAINER [1];
                    CONTAINER [0] = Parameter;

                    if (TYPE == VARIABLE_CLASS) {
                        Parameter->CLASS_DATA = (void *)STRING_VALUE;
                        ((CompiledClass *)CLASS_PTR)->LINKS++;
                    } else
                    if (TYPE == VARIABLE_DELEGATE) {
                        Parameter->CLASS_DATA = (void *)STRING_VALUE;
                        ((CompiledClass *)CLASS_PTR)->LINKS++;
                        Parameter->DELEGATE_DATA = (int)NUMBER_VALUE;
                    } else
                    if (TYPE == VARIABLE_ARRAY) {
                        Parameter->CLASS_DATA = (void *)STRING_VALUE;
                        ((Array *)CLASS_PTR)->LINKS++;
                    } else
                    if (TYPE == VARIABLE_STRING) {
                        Parameter->CLASS_DATA     = 0;
                        CONCEPT_STRING(Parameter) = STRING_VALUE;
                    } else
                        Parameter->NUMBER_DATA        = NUMBER_VALUE;

                    VariableDATA *Owner = (VariableDATA *)VAR_ALLOC(/*((CompiledClass *)CLASS_PTR)->PIF*/CCode->PIF);
                    //Owner->CALL_SET           = 0;
                    Owner->CLASS_DATA         = CLASS_PTR;
                    Owner->IS_PROPERTY_RESULT = 0;
                    Owner->LINKS       = 1;
                    Owner->TYPE        = VARIABLE_CLASS;
                    ((CompiledClass *)CLASS_PTR)->LINKS++;

                    try {
                        CCode->SetProperty(/*((CompiledClass *)CLASS_PTR)->PIF*/CCode->PIF,
                            index,
                            Owner,
                            &AE,
                            true,
                            1,
                            CONTAINER,
                            CCode->CLSID,
                            CCode->CLSID,
                            NULL);

                        FREE_VARIABLE(Owner);
                        FREE_VARIABLE(Parameter);
                        return(0);
                    } catch (VariableDATA *THROW_DATA) {
                        FREE_VARIABLE(THROW_DATA);
                        FREE_VARIABLE(Owner);
                        FREE_VARIABLE(Parameter);
                        return(-1);
                    }
                }
            }

            index = CCode->RELOCATIONS2 [index] - 1;
            if (index >= 0) {
                if ((CONTEXT) && (!m_type) && (!CONTEXT [index]) && (CM))
                    ((CompiledClass *)CLASS_PTR)->CreateVariable(index, CM);

                if (CONTEXT [index]) {
                    return(SetVariable(CONTEXT [index], TYPE, STRING_VALUE, NUMBER_VALUE));
                }
            } else {
                return(-1);
            }
        }
    }
    return(-1);
}

INTEGER Invoke(INTEGER INVOKE_TYPE, ...) {
    va_list ap;

    va_start(ap, INVOKE_TYPE);
    INTEGER result = INVOKE_SUCCESS;
    switch (INVOKE_TYPE) {
        case INVOKE_NEW_BUFFER:
            {
                INTEGER size  = va_arg(ap, INTEGER);
                char    **str = va_arg(ap, char **);
                *str = 0;
                if (size) {
                    *str = (char *)malloc(size);
                    if (!(*str)) {
                        result = CANNOT_INVOKE_IN_THIS_CASE;
                    }
                }
            }
            break;

        case INVOKE_DELETE_BUFFER:
            {
                char *str = va_arg(ap, char *);
                if (str) {
                    free(str);                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    //delete[] str;
                } else                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        {
                    result = CANNOT_INVOKE_IN_THIS_CASE;
                }
            }
            break;

        case INVOKE_SET_VARIABLE:
            {
                VariableDATA *vdata = va_arg(ap, VariableDATA *);
                INTEGER      type   = va_arg(ap, INTEGER);
                char         *str   = va_arg(ap, char *);
                NUMBER       nvalue = va_arg(ap, NUMBER);
                result = SetVariable(vdata, type, str, nvalue);
            }
            break;

        case INVOKE_GET_VARIABLE:
            {
                VariableDATA *vdata  = va_arg(ap, VariableDATA *);
                INTEGER      *type   = va_arg(ap, INTEGER *);
                char         **str   = va_arg(ap, char **);
                NUMBER       *nvalue = va_arg(ap, NUMBER *);
                result = GetVariable(vdata, type, str, nvalue);
            }
            break;

        case INVOKE_SET_CLASS_MEMBER:
            {
                void    *handle  = va_arg(ap, void *);
                char    *mname   = va_arg(ap, char *);
                INTEGER type     = va_arg(ap, INTEGER);
                char    *szvalue = va_arg(ap, char *);
                NUMBER  nvalue   = va_arg(ap, NUMBER);
                result = SetClassMember(handle, mname, type, szvalue, nvalue);
            }
            break;

        case INVOKE_GET_CLASS_MEMBER:
            {
                void    *handle   = va_arg(ap, void *);
                char    *mname    = va_arg(ap, char *);
                INTEGER *type     = va_arg(ap, INTEGER *);
                char    **szvalue = va_arg(ap, char **);
                NUMBER  *nvalue   = va_arg(ap, NUMBER *);
                result = GetClassMember(handle, mname, type, szvalue, nvalue);
                break;
            }
            break;

        case INVOKE_GET_CLASS_VARIABLE:
            {
                VariableDATA *handle = va_arg(ap, VariableDATA *);
                if ((handle) && (handle->TYPE == VARIABLE_CLASS)) {
                    char *mname = va_arg(ap, char *);
                    void **ptr  = va_arg(ap, void **);
                    result = GetClassMemberVariable(handle->CLASS_DATA, mname, ptr);
                }
                break;
            }
            break;

        case INVOKE_LOCK_VARIABLE:
            {
                VariableDATA *target = va_arg(ap, VariableDATA *);
                if (target) {
                    target->LINKS++;
                    // skipped by GC
                    if ((target->TYPE == VARIABLE_CLASS) || (target->TYPE == VARIABLE_DELEGATE))
                        ((CompiledClass *)target->CLASS_DATA)->reachable = 0x1C;
                    else
                    if (target->TYPE == VARIABLE_ARRAY)
                        ((CompiledClass *)target->CLASS_DATA)->reachable = 0x1C;
                } else {
                    result = INVALID_INVOKE_PARAMETER;
                }
            }
            break;

        case INVOKE_DYNAMIC_LOCK:
            {
                VariableDATA *target = va_arg(ap, VariableDATA *);
                if (target) {
                    if ((target->TYPE == VARIABLE_CLASS) || (target->TYPE == VARIABLE_DELEGATE)) {
                        ((CompiledClass *)target->CLASS_DATA)->LINKS++;
                        ((CompiledClass *)target->CLASS_DATA)->reachable = 0x1C;
                    } else
                    if (target->TYPE == VARIABLE_ARRAY) {
                        ((Array *)target->CLASS_DATA)->LINKS++;
                        ((CompiledClass *)target->CLASS_DATA)->reachable = 0x1C;
                    } else {
                        result = INVALID_INVOKE_PARAMETER;
                    }
                } else {
                    result = INVALID_INVOKE_PARAMETER;
                }
            }
            break;

        case INVOKE_CREATE_VARIABLE:
            {
                VariableDATA **target = va_arg(ap, VariableDATA * *);
                *target = (VariableDATA *)VAR_ALLOC(0);
                //(*target)->CALL_SET           = 0;
                //(*target)->CLASS_DATA         = 0;
                (*target)->IS_PROPERTY_RESULT = 0;
                (*target)->LINKS       = 1;
                (*target)->NUMBER_DATA = 0;
                (*target)->TYPE        = VARIABLE_NUMBER;
            }
            break;

        case INVOKE_FREE_VARIABLE:
            {
                VariableDATA *target = va_arg(ap, VariableDATA *);
                if (target) {
                    FREE_VARIABLE(target);
                } else {
                    result = INVALID_INVOKE_PARAMETER;
                }
            }
            break;

        case INVOKE_FREE_VARIABLE_REFERENCE:
            {
                VariableDATA *target = va_arg(ap, VariableDATA *);
                if (target) {
                    VAR_FREE(target);
                } else {
                    result = INVALID_INVOKE_PARAMETER;
                }
            }
            break;

        case INVOKE_CREATE_ARRAY:
            {
                VariableDATA *target = va_arg(ap, VariableDATA *);
                if (target) {
                    CLASS_CHECK(target);
                    target->TYPE       = VARIABLE_ARRAY;
                    PIFAlizator *PIF = (PIFAlizator *)POOL_CONTEXT(target);
                    target->CLASS_DATA = new(AllocArray(PIF)) Array(PIF);
                } else
                    result = INVALID_INVOKE_PARAMETER;
            }
            break;

        case INVOKE_CREATE_OBJECT:
            {
                PIFAlizator  *pif        = va_arg(ap, PIFAlizator *);
                VariableDATA *target     = va_arg(ap, VariableDATA *);
                char         *class_name = va_arg(ap, char *);
                if (!target) {
                    result = INVALID_INVOKE_PARAMETER;
                    break;
                }

                CLASS_CHECK(target);
                target->TYPE = VARIABLE_NUMBER;

                int       class_count = pif->ClassList->Count();
                int       clsid       = -1;
                ClassCode *CC         = 0;
                for (int i = 0; i < class_count; i++) {
                    CC = pif->StaticClassList[i];//(ClassCode *)pif->ClassList->Item(i);
                    if (CC->NAME == class_name) {
                        clsid = i;
                        break;
                    }
                }
                if (clsid == -1) {
                    result = INVALID_INVOKE_PARAMETER;
                    break;
                }

                // am clsid ... mai trebuie sa instantziez clasa ...
                ParamList FORMAL_PARAM;
                FORMAL_PARAM.COUNT       = 0;
                FORMAL_PARAM.PARAM_INDEX = 0;

                VariableDATA dummyVD [1];

                target->TYPE       = VARIABLE_CLASS;
                target->CLASS_DATA = CC->CreateInstance(pif, target, 0, &FORMAL_PARAM, (VariableDATA **)&dummyVD, NULL);
                if (!target->CLASS_DATA) {
                    target->TYPE = VARIABLE_NUMBER;
                    result       = INVALID_INVOKE_PARAMETER;
                    break;
                }

                target->TYPE = VARIABLE_CLASS;
            }
            break;

        case INVOKE_ARRAY_VARIABLE:
            {
                VariableDATA *target = va_arg(ap, VariableDATA *);
                INTEGER      index   = va_arg(ap, INTEGER);
                VariableDATA **res   = va_arg(ap, VariableDATA * *);
                if (!target) {
                    result = INVALID_INVOKE_PARAMETER;
                    break;
                }
                if ((target->TYPE == VARIABLE_ARRAY) && (target->CLASS_DATA)) {
                    // TO DO !
                    *res = ((Array *)target->CLASS_DATA)->ModuleGet(index);
                } else {
                    result = INVALID_INVOKE_PARAMETER;
                }
            }
            break;

        case INVOKE_ARRAY_VARIABLE_BY_KEY:
            {
                VariableDATA *target = va_arg(ap, VariableDATA *);
                char         *key    = va_arg(ap, char *);
                VariableDATA **res   = va_arg(ap, VariableDATA * *);
                if (!target) {
                    result = INVALID_INVOKE_PARAMETER;
                    break;
                }
                if ((target->TYPE == VARIABLE_ARRAY) && (target->CLASS_DATA)) {
                    *res = ((Array *)target->CLASS_DATA)->ModuleGet(key);
                } else {
                    result = INVALID_INVOKE_PARAMETER;
                }
            }
            break;

        case INVOKE_GET_ARRAY_ELEMENT:
            {
                VariableDATA *target = va_arg(ap, VariableDATA *);
                INTEGER      index   = va_arg(ap, INTEGER);
                void         *res    = 0;
                if (!target) {
                    result = INVALID_INVOKE_PARAMETER;
                    break;
                }
                if ((target->TYPE == VARIABLE_ARRAY) && (target->CLASS_DATA)) {
                    // TO DO !
                    res = ((Array *)target->CLASS_DATA)->ModuleGet(index);
                    INTEGER *type = va_arg(ap, INTEGER *);
                    char    **str = va_arg(ap, char **);
                    NUMBER  *nr   = va_arg(ap, double *);
                    GetVariable((VariableDATA *)res, type, str, nr);
                } else {
                    result = INVALID_INVOKE_PARAMETER;
                }
            }
            break;

        case INVOKE_GET_ARRAY_ELEMENT_BY_KEY:
            {
                VariableDATA *target = va_arg(ap, VariableDATA *);
                char         *key    = va_arg(ap, char *);
                void         *res    = 0;
                if (!target) {
                    result = INVALID_INVOKE_PARAMETER;
                    break;
                }
                if ((target->TYPE == VARIABLE_ARRAY) && (target->CLASS_DATA)) {
                    res = ((Array *)target->CLASS_DATA)->ModuleGet(key);
                    INTEGER *type = va_arg(ap, INTEGER *);
                    char    **str = va_arg(ap, char **);
                    NUMBER  *nr   = va_arg(ap, double *);
                    GetVariable((VariableDATA *)res, type, str, nr);
                } else {
                    result = INVALID_INVOKE_PARAMETER;
                }
            }
            break;

        case INVOKE_SET_ARRAY_ELEMENT:
            {
                VariableDATA *target = va_arg(ap, VariableDATA *);
                INTEGER      index   = va_arg(ap, INTEGER);
                void         *res    = 0;
                if (!target) {
                    result = INVALID_INVOKE_PARAMETER;
                    break;
                }
                if ((target->TYPE == VARIABLE_ARRAY) && (target->CLASS_DATA)) {
                    res = ((Array *)target->CLASS_DATA)->ModuleGet(index);

                    INTEGER type = va_arg(ap, INTEGER);
                    char    *str = va_arg(ap, char *);
                    NUMBER  nr   = va_arg(ap, double);
                    SetVariable((VariableDATA *)res, type, str, nr);
                } else {
                    result = INVALID_INVOKE_PARAMETER;
                }
            }
            break;

        case INVOKE_SET_ARRAY_ELEMENT_BY_KEY:
            {
                VariableDATA *target = va_arg(ap, VariableDATA *);
                char         *key    = va_arg(ap, char *);
                void         *res    = 0;
                if (!target) {
                    result = INVALID_INVOKE_PARAMETER;
                    break;
                }
                if ((target->TYPE == VARIABLE_ARRAY) && (target->CLASS_DATA)) {
                    res = ((Array *)target->CLASS_DATA)->ModuleGet(key);

                    INTEGER type = va_arg(ap, INTEGER);
                    char    *str = va_arg(ap, char *);
                    NUMBER  nr   = va_arg(ap, double);
                    SetVariable((VariableDATA *)res, type, str, nr);
                } else {
                    result = INVALID_INVOKE_PARAMETER;
                }
            }
            break;

        case INVOKE_GET_ARRAY_KEY:
            {
                VariableDATA *target = va_arg(ap, VariableDATA *);
                INTEGER      index   = va_arg(ap, INTEGER);
                char         **key   = va_arg(ap, char **);
                if (!target) {
                    result = INVALID_INVOKE_PARAMETER;
                    break;
                }
                if ((target->TYPE == VARIABLE_ARRAY) && (target->CLASS_DATA)) {
                    *key = ((Array *)target->CLASS_DATA)->GetKey(index);
                } else {
                    result = INVALID_INVOKE_PARAMETER;
                }
            }
            break;

        case INVOKE_ARRAY_ELEMENT_IS_SET:
            {
                VariableDATA *target = va_arg(ap, VariableDATA *);
                INTEGER      index   = va_arg(ap, INTEGER);
                char         *key    = va_arg(ap, char *);
                if (!target) {
                    result = INVALID_INVOKE_PARAMETER;
                    break;
                }
                if ((target->TYPE == VARIABLE_ARRAY) && (target->CLASS_DATA)) {
                    if (index >= 0) {
                        result = ((Array *)target->CLASS_DATA)->FindIndex(index) != -1;
                    } else {
                        result = ((Array *)target->CLASS_DATA)->FindKey(key) != -1;
                    }
                } else {
                    result = INVALID_INVOKE_PARAMETER;
                }
            }
            break;

        case INVOKE_GET_ARRAY_INDEX:
            {
                VariableDATA *target = va_arg(ap, VariableDATA *);
                char         *key    = va_arg(ap, char *);
                INTEGER      *index  = va_arg(ap, INTEGER *);
                if (!target) {
                    result = INVALID_INVOKE_PARAMETER;
                    break;
                }
                if ((target->TYPE == VARIABLE_ARRAY) && (target->CLASS_DATA)) {
                    *index = ((Array *)target->CLASS_DATA)->GetIndex(key);
                } else {
                    result = INVALID_INVOKE_PARAMETER;
                }
            }
            break;

        case INVOKE_GET_ARRAY_COUNT:
            {
                VariableDATA *target = va_arg(ap, VariableDATA *);
                if (!target) {
                    result = INVALID_INVOKE_PARAMETER;
                    break;
                }
                if ((target->TYPE == VARIABLE_ARRAY) && (target->CLASS_DATA)) {
                    result = ((Array *)target->CLASS_DATA)->Count();
                } else {
                    result = INVALID_INVOKE_PARAMETER;
                }
            }
            break;

        case INVOKE_CREATE_DELEGATE:
            {
                VariableDATA *target            = va_arg(ap, VariableDATA *);
                VariableDATA *deleg             = va_arg(ap, VariableDATA *);
                char *class_member_name         = va_arg(ap, char *);

                if (((target->TYPE == VARIABLE_CLASS) || (target->TYPE == VARIABLE_DELEGATE)) && (target->CLASS_DATA) && (deleg) && (class_member_name)) {
                    CLASS_CHECK(deleg);
                    int m_type=0;
                    INTEGER index  = ((CompiledClass *)target->CLASS_DATA)->_Class->HasMemberInCompiled(class_member_name, &m_type);
                    if (index) {
                        index--;
                        ClassMember *CM = index < ((CompiledClass *)target->CLASS_DATA)->_Class->pMEMBERS_COUNT ? ((CompiledClass *)target->CLASS_DATA)->_Class->pMEMBERS[index] : 0;

                        if ((CM) && (CM->IS_FUNCTION)) {
                            index = ((CompiledClass *)target->CLASS_DATA)->_Class->GetAbsoluteMemberID(index + 1) - 1;
                            if (index>=0) {
                                deleg->TYPE=VARIABLE_DELEGATE;
                                deleg->DELEGATE_DATA=index;
                                deleg->CLASS_DATA=target->CLASS_DATA;
                                ((CompiledClass *)target->CLASS_DATA)->LINKS++;
                            } else
                                result = INVALID_INVOKE_PARAMETER;
                        } else
                            result = INVALID_INVOKE_PARAMETER;
                    } else
                        result = INVALID_INVOKE_PARAMETER;
                } else
                    result = INVALID_INVOKE_PARAMETER;
            }
            break;

        case INVOKE_CALL_DELEGATE:
        case INVOKE_CALL_DELEGATE_THREAD:
        case INVOKE_CALL_DELEGATE_THREAD_SAFE:
            {
                VariableDATA *target            = va_arg(ap, VariableDATA *);
                VariableDATA **SENDER_RESULT    = va_arg(ap, VariableDATA * *);
                VariableDATA **SENDER_EXCEPTION = va_arg(ap, VariableDATA * *);

                if ((!target) || (target->TYPE!=VARIABLE_DELEGATE) || (!target->CLASS_DATA)) {
                    result = INVALID_INVOKE_PARAMETER;
                    break;
                }
                PIFAlizator *PIF=((CompiledClass *)target->CLASS_DATA)->_Class->PIF;
#ifdef SIMPLE_MULTI_THREADING
                char thread_created=0;
                if (((INVOKE_TYPE == INVOKE_CALL_DELEGATE_THREAD) || (INVOKE_TYPE == INVOKE_CALL_DELEGATE_THREAD_SAFE)) && (target->TYPE == VARIABLE_DELEGATE) && (target->CLASS_DATA)) {
                    NEW_THREAD
                    thread_created=1;
                }
#endif
                if ((PIF) && (INVOKE_TYPE == INVOKE_CALL_DELEGATE_THREAD_SAFE)) {
                    semp(PIF->DelegateLock);
                }

                if ((target->TYPE == VARIABLE_DELEGATE) && (target->CLASS_DATA)) {
                    INTEGER      parameters = va_arg(ap, INTEGER);
                    VariableDATA **CTX      = 0;
                    bool         is_list    = false;
                    if (parameters == -1) {
                        CTX = va_arg(ap, VariableDATA * *);
                        int i = 0;
                        while (CTX [i++]) {
                        }
                        parameters = i - 1;
                        is_list    = true;
                    } else {
                        CTX = parameters ? (VariableDATA **)FAST_MALLOC(sizeof(VariableDATA*) * parameters) : (VariableDATA **)0/*-1*/;
                    }
                    ParamList FORMAL_PARAM;
                    FORMAL_PARAM.COUNT       = parameters;
                    FORMAL_PARAM.PARAM_INDEX = parameters ? DELTA_REF(&FORMAL_PARAM, new INTEGER [parameters]) : 0;

                    if (is_list) {
                        for (int i = 0; i < parameters; i++) {
                            DELTA_UNREF(&FORMAL_PARAM, FORMAL_PARAM.PARAM_INDEX) [i] = i + 1;
                            CTX [i]->LINKS++;
                        }
                    } else {
                        for (int i = 0; i < parameters; i++) {
                            DELTA_UNREF(&FORMAL_PARAM, FORMAL_PARAM.PARAM_INDEX) [i] = i + 1;
                            INTEGER type = va_arg(ap, INTEGER);
                            if (type == VARIABLE_UNDEFINED) {
                                CTX [i] = va_arg(ap, VariableDATA *);
                                CTX [i]->LINKS++;
                            } else {
                                char   *str = va_arg(ap, char *);
                                NUMBER nr   = va_arg(ap, double);

                                CTX [i] = (VariableDATA *)VAR_ALLOC(PIF);
                                //CTX [i]->CALL_SET           = 0;
                                //CTX [i]->CLASS_DATA         = 0;
                                CTX [i]->TYPE = VARIABLE_NUMBER;
                                CTX [i]->IS_PROPERTY_RESULT = 0;
                                CTX [i]->LINKS       = 1;
                                CTX [i]->NUMBER_DATA = 0;
                                SetVariable(CTX [i], type, str, nr);
                            }
                        }
                    }
                    VariableDATA *RESULT = target;
                    VariableDATA *lOwner = 0;

                    try {
                        lOwner = (VariableDATA *)VAR_ALLOC(PIF);
                        //lOwner->ACCESS=ACCESS_LOCAL;
                        //lOwner->CALL_SET           = 0;
                        lOwner->CLASS_DATA         = target->CLASS_DATA;
                        //lOwner->EXTRA_DATA         = 0;
                        lOwner->IS_PROPERTY_RESULT = 0;
                        lOwner->LINKS       = 1;
                        //lOwner->NUMBER_DATA = 0;
                        lOwner->TYPE        = VARIABLE_CLASS;

                        ((CompiledClass *)RESULT->CLASS_DATA)->LINKS++;
                        *SENDER_RESULT = ((CompiledClass *)RESULT->CLASS_DATA)->_Class->ExecuteMember(PIF,
                            (INTEGER)RESULT->DELEGATE_DATA,
                            lOwner,
                            0,
                            1,
                            &FORMAL_PARAM,
                            CTX,
                            0,
                            ((CompiledClass *)RESULT->CLASS_DATA)->_Class->CLSID,
                            ((CompiledClass *)RESULT->CLASS_DATA)->_Class->CLSID,
                            NULL);
                        FREE_VARIABLE(lOwner);
                    } catch (VariableDATA *LAST_THROW) {
                        FREE_VARIABLE(lOwner);
                        *SENDER_EXCEPTION = LAST_THROW;
                    }
/*#ifdef SIMPLE_MULTI_THREADING
                    if (INVOKE_TYPE == INVOKE_CALL_DELEGATE_THREAD) {
                        DONE_THREAD
                    }
#endif*/

                    if ((CTX) && ((intptr_t)CTX != -1)) {
                        for (int i = 0; i < parameters; i++) {
                            FREE_VARIABLE(CTX [i]);
                        }
                        if (!is_list) {
                            FAST_FREE(CTX);
                        }
                    }
                    if (FORMAL_PARAM.PARAM_INDEX) {
                        delete[] DELTA_UNREF(&FORMAL_PARAM, FORMAL_PARAM.PARAM_INDEX);
                    }
                } else {
                    result = INVALID_INVOKE_PARAMETER;
                }
#ifdef SIMPLE_MULTI_THREADING
                if (thread_created) {
                    DONE_THREAD
                }
#endif
                if ((PIF) && (INVOKE_TYPE == INVOKE_CALL_DELEGATE_THREAD_SAFE)) {
                    semv(PIF->DelegateLock);
                }
            }
            break;
        case INVOKE_EXTERNAL_THREADING:
            {
                VariableDATA *target = va_arg(ap, VariableDATA *);
                INTEGER i = va_arg(ap, INTEGER);
                if ((!target) || ((target->TYPE!=VARIABLE_DELEGATE) && (target->TYPE!=VARIABLE_CLASS)) || (!target->CLASS_DATA)) {
                    result = INVALID_INVOKE_PARAMETER;
                    break;
                }
                PIFAlizator *PIF=((CompiledClass *)target->CLASS_DATA)->_Class->PIF;
                NEW_THREAD
            }
            break;
        case INVOKE_THREAD_LOCK:
            {
                VariableDATA *target = va_arg(ap, VariableDATA *);
                INTEGER i = va_arg(ap, INTEGER);
                if ((!target) || ((target->TYPE!=VARIABLE_DELEGATE) && (target->TYPE!=VARIABLE_CLASS)) || (!target->CLASS_DATA)) {
                    result = INVALID_INVOKE_PARAMETER;
                    break;
                }

                PIFAlizator *PIF=((CompiledClass *)target->CLASS_DATA)->_Class->PIF;
                if (!PIF) {
                    result = CANNOT_INVOKE_IN_THIS_CASE;
                    break;
                }
                if (i) {
                    semp(PIF->DelegateLock);
                } else {
                    semv(PIF->DelegateLock);
                }
            }
            break;

        case INVOKE_COUNT_DELEGATE_PARAMS:
            {
                VariableDATA *target = va_arg(ap, VariableDATA *);
                if (!target) {
                    result = INVALID_INVOKE_PARAMETER;
                    break;
                }
                if ((target->TYPE == VARIABLE_DELEGATE) && (target->CLASS_DATA)) {
                    int mid = ((CompiledClass *)target->CLASS_DATA)->_Class->RELOCATIONS [(intptr_t)target->DELEGATE_DATA];
                    ClassMember *CM = mid ? ((CompiledClass *)target->CLASS_DATA)->_Class->pMEMBERS [mid - 1] : 0;
                    if (CM) {
                        result = CM->PARAMETERS_COUNT;
                    } else {
                        result = CANNOT_INVOKE_IN_THIS_CASE;
                    }
                } else {
                    result = INVALID_INVOKE_PARAMETER;
                }
            }
            break;

        case INVOKE_COUNT_DELEGATE_PARAMS2:
            {
                VariableDATA *target = va_arg(ap, VariableDATA *);
                INTEGER *params = va_arg(ap, INTEGER *);
                INTEGER *min_params = va_arg(ap, INTEGER *);
                if (!target) {
                    result = INVALID_INVOKE_PARAMETER;
                    break;
                }
                if ((target->TYPE == VARIABLE_DELEGATE) && (target->CLASS_DATA)) {
                    int mid = ((CompiledClass *)target->CLASS_DATA)->_Class->RELOCATIONS [(intptr_t)target->DELEGATE_DATA];
                    ClassMember *CM = mid ? ((CompiledClass *)target->CLASS_DATA)->_Class->pMEMBERS [mid - 1] : 0;
                    if (CM) {
                        *params = CM->PARAMETERS_COUNT;
                        *min_params = CM->MUST_PARAMETERS_COUNT;
                    } else {
                        result = CANNOT_INVOKE_IN_THIS_CASE;
                    }
                } else {
                    result = INVALID_INVOKE_PARAMETER;
                }
            }
            break;

        case INVOKE_GET_DELEGATE_NAMES:
            {
                VariableDATA *target       = va_arg(ap, VariableDATA *);
                char         **class_name  = va_arg(ap, char **);
                char         **member_name = va_arg(ap, char **);
                if (!target) {
                    result = INVALID_INVOKE_PARAMETER;
                    break;
                }
                if ((target->TYPE == VARIABLE_DELEGATE) && (target->CLASS_DATA)) {
                    int mid = ((CompiledClass *)target->CLASS_DATA)->_Class->RELOCATIONS [(intptr_t)target->DELEGATE_DATA];
                    ClassMember *CM = mid ? ((CompiledClass *)target->CLASS_DATA)->_Class->pMEMBERS [mid - 1] : 0;
                    if (CM) {
                        *class_name  = ((CompiledClass *)target->CLASS_DATA)->_Class->NAME.c_str();
                        *member_name = CM->NAME;
                    } else {
                        result = CANNOT_INVOKE_IN_THIS_CASE;
                    }
                } else {
                    result = INVALID_INVOKE_PARAMETER;
                }
            }
            break;

        case INVOKE_DEFINE_CLASS:
            {
                PIFAlizator *pif = va_arg(ap, PIFAlizator *);

                char *classname = va_arg(ap, char *);

                if (pif->ClassExists(classname)) {
                    result = INVALID_INVOKE_PARAMETER;
                    break;
                }

                ClassCode *CC = new ClassCode(classname, pif);

                char *member_name = va_arg(ap, char *);
                char is_array     = 0;
                while (member_name) {
                    size_t        len = strlen(member_name);
                    AnsiString name;
                    if ((member_name [len - 2] == '[') && (member_name [len - 1] == ']')) {
                        is_array = 1;
                        name.LoadBuffer(member_name, len - 2);
                    } else {
                        name     = member_name;
                        is_array = 0;
                    }

                    char *ref_mname=pif->CheckMember( name);
                    ClassMember *CM = CC->AddMember(ref_mname, 0, (char *)"STATIC_MODULE", ACCESS_PUBLIC, true);
                    if (is_array) {
                        CM->VD->TYPE = VARIABLE_ARRAY;
                    }
                    member_name=va_arg(ap,char *);
                }
                CC->NEEDED        = 0;
                CC->DEFINED_LEVEL = -1;
                pif->ClassList->Add(CC, DATA_CLASS_CODE);
#ifdef CACHED_CLASSES
                unsigned int key=murmur_hash(CC->NAME.c_str(), CC->NAME.Length());
                pif->CachedClasses[key]=pif->ClassList.Count();
#endif
                pif->SyncClassList();
            }
            break;

        case INVOKE_DEFINE_CONSTANT:
            {
                PIFAlizator *pif      = va_arg(ap, PIFAlizator *);
                char        *CONSTANT = va_arg(ap, char *);
                char        *value    = va_arg(ap, char *);

                if ((pif) && (CONSTANT) && (value)) {
                    int ct_count           = pif->ConstantList.Count();
                    VariableDESCRIPTOR *VD = 0;
                    for (int i = 0; i < ct_count; i++) {
                        VariableDESCRIPTOR *VTEMP = (VariableDESCRIPTOR *)pif->ConstantList.Item(i);
                        if (VTEMP->name == CONSTANT) {
                            VD = VTEMP;
                            break;
                        }
                    }
                    if (!VD) {
                        VD         = new VariableDESCRIPTOR;
                        VD->BY_REF = 0;
                        pif->ConstantList.Add(VD, DATA_VAR_DESCRIPTOR);
                    }
                    VD->name  = CONSTANT;
                    VD->value = value;
                    VD->USED  = -1;
                } else {
                    result = INVALID_INVOKE_PARAMETER;
                }
            }
            break;

        case INVOKE_GET_APPLICATION_INFO:
            {
                PIFAlizator *pif         = va_arg(ap, PIFAlizator *);
                int         *pipein      = va_arg(ap, int *);
                int         *pipeout     = va_arg(ap, int *);
                int         *apid        = va_arg(ap, int *);
                int         *parent_apid = va_arg(ap, int *);

                if ((pif) && (apid) && (pipein) && (pipeout) && (parent_apid)) {
                    *pipein      = pif->pipe_read;
                    *pipeout     = pif->pipe_write;
                    *apid        = pif->apid;
                    *parent_apid = pif->parent_apid;
                }
            }
            break;

        case INVOKE_GET_SERIAL_CLASS:
            {
                CompiledClass *CC             = va_arg(ap, CompiledClass *);
                int           max_members     = va_arg(ap, int);
                char          **class_name    = va_arg(ap, char **);
                char          **pmembers      = va_arg(ap, char **);
                unsigned char *flags          = va_arg(ap, unsigned char *);
                char          *access         = va_arg(ap, char *);
                char          *types          = va_arg(ap, char *);
                char          **szValue       = va_arg(ap, char **);
                double        *n_data         = va_arg(ap, double *);
                void          **class_data    = va_arg(ap, void **);
                void          **variable_data = va_arg(ap, void **);

                if (!CC) {
                    result = CANNOT_INVOKE_INTERFACE;
                    break;
                }
                *class_name = CC->_Class->NAME.c_str();
                result      = CC->_Class->GetSerialMembers(CC, max_members, pmembers, flags, access, types, szValue, n_data, class_data, variable_data);
            }
            break;

        case INVOKE_GET_MEMBER_FROM_ID:
            {
                CompiledClass *CC       = va_arg(ap, CompiledClass *);
                intptr_t          member_id = va_arg(ap, intptr_t);
                char          **member  = va_arg(ap, char **);
                if ((!CC) || (member_id<0)) {
                    result = CANNOT_INVOKE_INTERFACE;
                    break;
                }
                int mid = CC->_Class->RELOCATIONS [member_id];
                ClassMember *CM = mid ? CC->_Class->pMEMBERS [mid - 1] : 0;
                if (CM) {
                    *member = CM->NAME;
                } else {
                    result = CANNOT_INVOKE_IN_THIS_CASE;
                }
            }
            break;

        case INVOKE_HAS_MEMBER:
            {
                CompiledClass *CC     = va_arg(ap, CompiledClass *);
                char          *member = va_arg(ap, char *);
                if (!CC) {
                    result = CANNOT_INVOKE_INTERFACE;
                    break;
                }

                if (CC->_Class->HasMember(member)) {
                    result = 1;
                } else {
                    result = 0;
                }
            }
            break;

        case INVOKE_OBJECT_LINKS:
            {
                VariableDATA *target = va_arg(ap, VariableDATA *);
                if (!target) {
                    result = INVALID_INVOKE_PARAMETER;
                    break;
                }
                if (target->CLASS_DATA) {
                    if ((target->TYPE == VARIABLE_CLASS) || (target->TYPE == VARIABLE_DELEGATE)) {
                        result = ((CompiledClass *)target->CLASS_DATA)->LINKS;
                    } else
                    if (target->TYPE == VARIABLE_ARRAY) {
                        result = ((Array *)target->CLASS_DATA)->LINKS;
                    }
                }
            }
            break;

        case INVOKE_VAR_LINKS:
            {
                VariableDATA *target = va_arg(ap, VariableDATA *);
                if (target) {
                    result = target->LINKS;
                }
            }
            break;

        case INVOKE_CLI_ARGUMENTS:
            {
                intptr_t *argc   = va_arg(ap, intptr_t *);
                char ***argv = va_arg(ap, char ***);
                *argv = PIFAlizator::argv;
                *argc = PIFAlizator::argc;
            }
            break;

        case INVOKE_SORT_ARRAY_BY_KEYS:
            {
                Array *target = va_arg(ap, Array *);
                Array **res   = va_arg(ap, Array * *);

                if (target) {
                    *res = target->SortArrayElementsByKey();
                    if (*res)
                        (*res)->LINKS = 0;
                } else {
                    result = INVALID_INVOKE_PARAMETER;
                }
            }
            break;

        case INVOKE_CHECK_POINT:
            {
                int status = va_arg(ap, int);
                if (PIFAlizator::CheckPoint) {
                    PIFAlizator::CheckPoint(status);
                } else {
                    result = CANNOT_INVOKE_INTERFACE;
                }
            }
            break;

        case INVOKE_GET_USERDATA:
            {
                PIFAlizator *pif   = va_arg(ap, PIFAlizator *);
                void        **data = va_arg(ap, void **);
                if (!pif) {
                    result = INVALID_INVOKE_PARAMETER;
                } else {
                    *data = pif->out->userdata;
                }
            }
            break;

        case INVOKE_GET_THREAD_DATA:
            {
                void **data = va_arg(ap, void **);
                *data = GetStaticData();
            }
            break;

        case INVOKE_SET_THREAD_DATA:
            {
                void *data = va_arg(ap, void *);
                SetStaticData(data);
            }
            break;

        case INVOKE_PRINT:
            {
                PIFAlizator *PIF  = va_arg(ap, PIFAlizator *);
                char        *what = va_arg(ap, char *);
                INTEGER     len   = va_arg(ap, INTEGER);
                PIF->out->Print(what, len);
            }
            break;

        case INVOKE_GET_PROTO_SOCKET:
            {
                PIFAlizator *PIF  = va_arg(ap, PIFAlizator *);
                if ((PIF->direct_pipe>0) && (eof(PIF->direct_pipe)==0)) {
#ifdef _WIN32
                    long new_sock=0;
                    if (read(PIF->direct_pipe, &new_sock, sizeof(new_sock))==sizeof(new_sock)) {
#else
                    int new_sock=0;
                    if (sock_fd_read(PIF->direct_pipe, &new_sock)>=0) {
#endif
                        if (new_sock>0) {
                            PIF->out->sock=new_sock;
                            PIFAlizator::refSOCKET=new_sock;
                        }
                    }
                }
                result = PIF->out->sock;
            }
            break;

        case INVOKE_HIBERNATE:
            {
                PIFAlizator *pif  = va_arg(ap, PIFAlizator *);
                char        *flag = va_arg(ap, char *);
                if (!pif) {
                    result = INVALID_INVOKE_PARAMETER;
                } else {
                    if (*flag) {
                        pif->Hibernate();
                    } else {
                        //pif->HibernateWake();
                    }
                }
            }
            break;

        case INVOKE_DELEGATE_BYTECODE:
            {
                VariableDATA *target            = va_arg(ap, VariableDATA *);
                void **CODE                     = va_arg(ap, void **);
                INTEGER *len                    = va_arg(ap, INTEGER *);

                void **DATA                     = va_arg(ap, void **);
                INTEGER *dlen                    = va_arg(ap, INTEGER *);

                if ((!target) || (target->TYPE!=VARIABLE_DELEGATE) || (!target->CLASS_DATA)) {
                    result = INVALID_INVOKE_PARAMETER;
                    break;
                }
                ClassCode *CC=((CompiledClass *)target->CLASS_DATA)->_Class;
                int i=(int)target->DELEGATE_DATA;
                int relocation  = CC->RELOCATIONS [i];
                ClassMember *pMEMBER_i = relocation ? CC->pMEMBERS [relocation - 1] : 0;
                if ((pMEMBER_i) && (pMEMBER_i->OPTIMIZER)) {
                    OptimizedElement *OElist = ((Optimizer *)pMEMBER_i->OPTIMIZER)->CODE;
                    ParamList *Parameters = ((Optimizer *)pMEMBER_i->OPTIMIZER)->PARAMS;
                    int count = ((Optimizer *)pMEMBER_i->OPTIMIZER)->codeCount;
                    *len = count;
                    TreeContainer *tc=(TreeContainer *)malloc(sizeof(TreeContainer)*count);
                    for (int it=0;it<count;it++) {
                        OptimizedElement *OE=&OElist[it];
                        TreeContainer *reftc=&tc[it];

                        reftc->Operator_ID = OE->Operator.ID;
                        reftc->Operator_TYPE = OE->Operator.TYPE;
                        reftc->OperandLeft_ID = OE->OperandLeft.ID;
                        reftc->OperandLeft_TYPE = OE->OperandLeft.TYPE;
                        reftc->OperandRight_ID = OE->OperandRight.ID;
                        reftc->OperandRight_TYPE= OE->OperandRight.TYPE;
                        if ((OE->Operator.TYPE == TYPE_OPERATOR) && (OE->Operator.ID == KEY_DLL_CALL) && (OE->OperandLeft.ID == STATIC_CLASS_DLL)) {
                            reftc->Function = OE->OperandRight._PARSE_DATA.c_str();
                        } else {
                            reftc->Function = 0;
                        }

                        if ((OE->Operator.TYPE == TYPE_OPERATOR) && ((OE->Operator.ID == KEY_DLL_CALL) || (OE->Operator.ID == KEY_SEL) || (OE->Operator.ID == KEY_NEW))) {
                            if ((OE->Operator.ID == KEY_SEL) && (OE->OperandLeft.ID == 1)) {
                                int i2=(int)OE->OperandRight.ID-1;
                                int relocation2 = CC->RELOCATIONS [i2];
                                ClassMember *pMEMBER_i2 = relocation2 ? CC->pMEMBERS [relocation2 - 1] : 0;
                                if (pMEMBER_i2)
                                    reftc->Function = pMEMBER_i2->NAME;
                            }
                            if (((OE->Operator.ID != KEY_NEW) || (OE->OperandLeft.ID != -1)) && (OE->OperandReserved.ID > 0))  {
                                ParamList *FORMAL_PARAMETERS = &Parameters[OE->OperandReserved.ID - 1];
                                reftc->ParametersCount = FORMAL_PARAMETERS->COUNT;
                                reftc->Parameters = DELTA_UNREF(FORMAL_PARAMETERS, FORMAL_PARAMETERS->PARAM_INDEX);
                            } else {
                                reftc->ParametersCount = -1;
                                reftc->Parameters = 0;
                            }
                        } else {
                            reftc->ParametersCount = -1;
                            reftc->Parameters = 0;
                        }
                        reftc->OperandReserved_ID = OE->OperandReserved.ID;
                        reftc->OperandReserved_TYPE = OE->OperandReserved.TYPE;
                        reftc->Result_ID = OE->Result_ID;
                    }
                    *CODE = tc;
                    count = ((Optimizer *)pMEMBER_i->OPTIMIZER)->dataCount;
                    VariableDESCRIPTOR *VDS = ((Optimizer *)pMEMBER_i->OPTIMIZER)->DATA;
                    *dlen = count;
                    TreeVD *tvd=(TreeVD *)malloc(sizeof(TreeVD)*count);
                    for (int it=0;it<count;it++) {
                        VariableDESCRIPTOR *VD = &VDS[it];
                        tvd[it].TYPE = VD->TYPE;
                        tvd[it].nValue = VD->nValue;
                        tvd[it].Value = VD->value.c_str();
                        tvd[it].Length = VD->value.Length();
                        tvd[it].IsRef = VD->BY_REF;
                    }
                    *DATA = tvd;
                } else {
                    *CODE = 0;
                    *len = 0;
                    *DATA = 0;
                    *dlen = 0;
                    result = INVALID_INVOKE_PARAMETER;
                }
            }
            break;

        case INVOKE_BYTECODE_FREE:
            {
                void *CODE                     = va_arg(ap, void *);
                void *DATA                     = va_arg(ap, void *);

                if (CODE)
                    free(CODE);
                if (DATA)
                    free(DATA);
            }
            break;

        case INVOKE_PROFILER:
            {
                PIFAlizator *pif      = va_arg(ap, PIFAlizator *);
                pif->PROFILE_DRIVEN   = va_arg(ap, char *);
            }
            break;

        case INVOKE_MULTITHREADED:
#ifdef SIMPLE_MULTI_THREADING
            result = 1;
#else
            result = 0;
#endif
            break;

        case INVOKE_GREENTHREAD:
            {
                VariableDATA *target            = va_arg(ap, VariableDATA *);
                GreenThreadCycle **cycle        = va_arg(ap, GreenThreadCycle **);
                *cycle = 0;
                if ((!target) || (target->TYPE!=VARIABLE_DELEGATE) || (!target->CLASS_DATA)) {
                    result = INVALID_INVOKE_PARAMETER;
                    break;
                }
                PIFAlizator *PIF=((CompiledClass *)target->CLASS_DATA)->_Class->PIF;
                if ((target->TYPE == VARIABLE_DELEGATE) && (target->CLASS_DATA)) {
                    try {
                        *cycle = ((CompiledClass *)target->CLASS_DATA)->_Class->CreateThread(PIF, (INTEGER)target->DELEGATE_DATA, target);
                    } catch (VariableDATA *LAST_THROW) {
                        FREE_VARIABLE(LAST_THROW);
                    }
                } else {
                    result = INVALID_INVOKE_PARAMETER;
                }
            }
            break;

        case INVOKE_FREEGREENTHREAD:
            {
                GreenThreadCycle *cycle        = va_arg(ap, GreenThreadCycle *);
                if (cycle)
                    ClassMember::DoneThread(cycle);
            }
            break;

        case INVOKE_GREENLOOP:
            {
                GreenThreadCycle *cycle        = va_arg(ap, GreenThreadCycle *);
                GreenThreadCycle *last         = va_arg(ap, GreenThreadCycle *);
                va_end(ap);
                if (cycle) {
                    cycle->LAST_THREAD = last;
                    return ConceptInterpreter::StacklessInterpret((PIFAlizator *)cycle->PIF, cycle);
                } else
                    return(CANNOT_INVOKE_INTERFACE);
            }
            break;

        case INVOKE_GREENLINK:
            {
                GreenThreadCycle *prev        = va_arg(ap, GreenThreadCycle *);
                GreenThreadCycle *current     = va_arg(ap, GreenThreadCycle *);
                if (prev) {
                    prev->NEXT = current;
                } else {
                    result = CANNOT_INVOKE_INTERFACE;
                }
            }
            break;

        case INVOKE_CURRENTGREENTHREAD:
            {
                GreenThreadCycle *cycle       = va_arg(ap, GreenThreadCycle *);
                GreenThreadCycle **current    = va_arg(ap, GreenThreadCycle **);
                if (cycle)
                    *current = (GreenThreadCycle *)cycle->CURRENT_THREAD;
                else
                    *current = 0;
            }
            break;

        case INVOKE_GREENINSERT:
            {
                GreenThreadCycle *cycle       = va_arg(ap, GreenThreadCycle *);
                GreenThreadCycle *current     = va_arg(ap, GreenThreadCycle *);
                if ((cycle) && (cycle->CURRENT_THREAD)) {
#ifdef SIMPLE_MULTI_THREADING
                    PIFAlizator *PIF=((CompiledClass *)((VariableDATA *)cycle->OWNER)->CLASS_DATA)->_Class->PIF;
                    semp(PIF->DelegateLock);
#endif

                    GreenThreadCycle *target = (GreenThreadCycle *)cycle->CURRENT_THREAD;
                    current->NEXT = target->NEXT;
                    target->NEXT = current;

#ifdef SIMPLE_MULTI_THREADING
                    semv(PIF->DelegateLock);
#endif
                } else
                    result = CANNOT_INVOKE_INTERFACE;
            }
            break;

        case INVOKE_SETSENDMESSAGEFUNCTION:
            {
                PIFAlizator *pif      = va_arg(ap, PIFAlizator *);
                SECURE_MESSAGE sm     = va_arg(ap, SECURE_MESSAGE);
                if (pif) {
                    if (pif->out)
                        pif->out->send_secure_message=sm;
                } else
                    result = INVALID_INVOKE_PARAMETER;
            }
            break;

        case INVOKE_GETGLOBALS:
            {
                PIFAlizator *pif      = va_arg(ap, PIFAlizator *);
                Array **arr           = va_arg(ap, Array **);
                if ((pif) && (arr)) {
                    if (!pif->var_globals) {
                        pif->var_globals=new(AllocArray(pif)) Array(pif);
                        // never to be deleted
                        pif->var_globals->LINKS = 2;
                    }
                    *arr = pif->var_globals;
                } else
                    result = INVALID_INVOKE_PARAMETER;
            }
            break;

        case INVOKE_GETID:
            {
                PIFAlizator *pif      = va_arg(ap, PIFAlizator *);
                if (pif)
                    return ++pif->IDGenerator;
                else
                    result = INVALID_INVOKE_PARAMETER;
            }
            break;

        case INVOKE_GETPROTODATA:
            {
                PIFAlizator *pif      = va_arg(ap, PIFAlizator *);
                int position          = va_arg(ap, int);
                void **data           = va_arg(ap, void **);
                if ((pif) && (data) && (position>=0) && (position<PDATA_ITEMS))
                    *data = pif->PDATA[position].data;
                else
                    result = INVALID_INVOKE_PARAMETER;
            }
            break;

        case INVOKE_SETPROTODATA:
            {
                PIFAlizator *pif      = va_arg(ap, PIFAlizator *);
                int position          = va_arg(ap, int);
                void *data           = va_arg(ap, void *);
                DESTROY_PROTO_DATA destroy_f = va_arg(ap, DESTROY_PROTO_DATA);
                if ((pif) && (data) && (position>=0) && (position<PDATA_ITEMS)) {
                    pif->PDATA[position].data = data;
                    pif->PDATA[position].destroy_func = destroy_f;
                } else
                    result = INVALID_INVOKE_PARAMETER;
            }
            break;

        case INVOKE_EXECUTECODE:
            {
                PIFAlizator *pif = va_arg(ap, PIFAlizator *);
                char *code = va_arg(ap, char *);
                int len = va_arg(ap, int);
                VariableDATA *res = va_arg(ap, VariableDATA *);
                if ((code) && (len > 0))
                    result = Concept_ExecuteBuffer(code, len, pif->INCLUDE_DIR.c_str(), pif->IMPORT_DIR.c_str(), pif->out->fprint, pif->out->sock, 0, NULL, NULL, pif->SERVER_PUBLIC_KEY, pif->SERVER_PRIVATE_KEY, pif->CLIENT_PUBLIC_KEY, pif->pipe_read, pif->pipe_write, pif->apid, pif->parent_apid, PIFAlizator::CheckPoint, res, pif->ModuleList, NULL);
                else
                    result = INVALID_INVOKE_PARAMETER;
            }
            break;

        case INVOKE_MEMBER_NAME:
            {
                PIFAlizator *pif = va_arg(ap, PIFAlizator *);
                INTEGER index = va_arg(ap, INTEGER);
                char **name = va_arg(ap, char **);
                if (name)
                    *name = 0;

                if ((name) && (index>0)) {
                    if (pif->parentPIF)
                        *name = ((PIFAlizator *)pif->parentPIF)->GeneralMembers.Item(index - 1);
                    else
                        *name = pif->GeneralMembers.Item(index - 1);
                } else
                    result = INVALID_INVOKE_PARAMETER;
            }
            break;

        case INVOKE_CLASS_NAME:
            {
                PIFAlizator *pif = va_arg(ap, PIFAlizator *);
                INTEGER index = va_arg(ap, INTEGER);
                char **name = va_arg(ap, char **);
                if (name)
                    *name = 0;

                if ((name) && (index>0) && (pif->ClassList) && (index <= pif->ClassList->Count())) {
                    ClassCode *CC = (ClassCode *)pif->StaticClassList[index - 1];//(ClassCode *)pif->ClassList->Item(index - 1);
                    if (CC)
                        *name = CC->NAME.c_str();
                } else
                    result = INVALID_INVOKE_PARAMETER;
            }
            break;

        case INVOKE_MEMBER_INFO:
            {
                VariableDATA *target            = va_arg(ap, VariableDATA *);
                char *class_member_name         = va_arg(ap, char *);
                char **mdef                     = va_arg(ap, char **);
                INTEGER *type                   = va_arg(ap, INTEGER *);
                INTEGER *mset                   = va_arg(ap, INTEGER *);
                INTEGER *mget                   = va_arg(ap, INTEGER *);
                INTEGER *macc                   = va_arg(ap, INTEGER *);

                if (mdef)
                    *mdef = 0;
                if (type)
                    *type = -1;
                if (mset)
                    *mset = -1;
                if (mget)
                    *mget = -1;
                if (macc)
                    *macc = -1;

                if (((target->TYPE == VARIABLE_CLASS) || (target->TYPE == VARIABLE_DELEGATE)) && (target->CLASS_DATA) && (class_member_name)) {
                    int m_type=0;
                    INTEGER index  = ((CompiledClass *)target->CLASS_DATA)->_Class->HasMemberInCompiled(class_member_name, &m_type);
                    if (index) {
                        index--;
                        ClassMember *CM = index < ((CompiledClass *)target->CLASS_DATA)->_Class->pMEMBERS_COUNT ? ((CompiledClass *)target->CLASS_DATA)->_Class->pMEMBERS[index] : 0;

                        if (CM)  {
                            if (mdef) {
                                if (CM->Defined_In)
                                    *mdef = ((ClassCode *)CM->Defined_In)->NAME.c_str();
                            }
                            if (type) {
                                *type = CM->IS_FUNCTION;
                                if (CM->IS_OPERATOR)
                                    *type = 0x11;
                            }
                            if (mget)
                                *mget = CM->MEMBER_GET;
                            if (mset)
                                *mset = CM->MEMBER_SET;
                            if (macc)
                                *macc = CM->ACCESS;
                        } else
                            result = INVALID_INVOKE_PARAMETER;
                    } else
                        result = INVALID_INVOKE_PARAMETER;
                } else
                    result = INVALID_INVOKE_PARAMETER;
            }
            break;

        case INVOKE_CREATE_VARIABLE_2:
            {
                PIFAlizator *pif = va_arg(ap, PIFAlizator *);
                VariableDATA **target = va_arg(ap, VariableDATA * *);
                *target = (VariableDATA *)VAR_ALLOC(pif);
                (*target)->IS_PROPERTY_RESULT = 0;
                (*target)->LINKS       = 1;
                (*target)->NUMBER_DATA = 0;
                (*target)->TYPE        = VARIABLE_NUMBER;
            }
            break;

        case INVOKE_GET_KEYS:
            {
                PIFAlizator *pif = va_arg(ap, PIFAlizator *);
                if (pif) {
                    char **SERVER_PUBLIC_KEY = va_arg(ap, char **);
                    char **SERVER_PRIVATE_KEY = va_arg(ap, char **);
                    char **CLIENT_PUBLIC_KEY = va_arg(ap, char **);
                    if (SERVER_PUBLIC_KEY)
                        *SERVER_PUBLIC_KEY = pif->SERVER_PUBLIC_KEY;
                    if (SERVER_PRIVATE_KEY)
                        *SERVER_PRIVATE_KEY = pif->SERVER_PRIVATE_KEY;
                    if (CLIENT_PUBLIC_KEY)
                        *CLIENT_PUBLIC_KEY = pif->CLIENT_PUBLIC_KEY;
                } else
                    result = INVALID_INVOKE_PARAMETER;
            }
            break;

        case INVOKE_PROFILE_MEMORY:
            {
                PIFAlizator *PIF = va_arg(ap, PIFAlizator *);
                VariableDATA *RESULT = va_arg(ap, VariableDATA *);
                CLASS_CHECK(RESULT);
                RESULT->TYPE = VARIABLE_ARRAY;
                RESULT->CLASS_DATA = new(AllocArray(PIF)) Array(PIF);
                return GetMemoryStatistics(PIF, RESULT->CLASS_DATA);
            }
            break;

        default:
            va_end(ap);
            return(CANNOT_INVOKE_INTERFACE);
    }
    va_end(ap);
    return(result);
}

SYS_INT LinkFunction(char *FUNCTION_NAME, AnsiList *TARGET, void **CACHED_hDLL) {
#ifdef CACHE_MEMBERS
    unsigned int key=murmur_hash(FUNCTION_NAME, strlen(FUNCTION_NAME));
    SYS_INT val=FunctionMap[key];
    if (val)
        return val;
#endif

    AnsiString PREFIX  = (char *)"CONCEPT__";
    AnsiString PREFIX2 = (char *)"CONCEPT_";
    int        count   = TARGET->Count();

    if (CACHED_hDLL) {
#ifdef _WIN32
        EXTERNAL_CALL _PROC_ADR = (EXTERNAL_CALL)GetProcAddress((HMODULE)*CACHED_hDLL, PREFIX2 + FUNCTION_NAME);
#else
        EXTERNAL_CALL _PROC_ADR = (EXTERNAL_CALL)dlsym((HMODULE)*CACHED_hDLL, PREFIX2 + FUNCTION_NAME);
#endif
        if (!_PROC_ADR) {
#ifdef _WIN32
            _PROC_ADR = (EXTERNAL_CALL)GetProcAddress((HMODULE)*CACHED_hDLL, PREFIX + FUNCTION_NAME);
#else
            _PROC_ADR = (EXTERNAL_CALL)dlsym((HMODULE)*CACHED_hDLL, PREFIX + FUNCTION_NAME);
#endif
        }

        if (_PROC_ADR) {
#ifdef CACHE_MEMBERS
            FunctionMap[key]=(SYS_INT)_PROC_ADR;
#endif
            return((SYS_INT)_PROC_ADR);
        }
    }
    for (INTEGER i = 0; i < count; i++) {
        HMODULE hDLL = (HMODULE)TARGET->Item(i);
#ifdef _WIN32
        EXTERNAL_CALL _PROC_ADR = (EXTERNAL_CALL)GetProcAddress(hDLL, PREFIX2 + FUNCTION_NAME);
#else
        EXTERNAL_CALL _PROC_ADR = (EXTERNAL_CALL)dlsym(hDLL, PREFIX2 + FUNCTION_NAME);
#endif
        if (!_PROC_ADR) {
#ifdef _WIN32
            _PROC_ADR = (EXTERNAL_CALL)GetProcAddress(hDLL, PREFIX + FUNCTION_NAME);
#else
            _PROC_ADR = (EXTERNAL_CALL)dlsym(hDLL, PREFIX + FUNCTION_NAME);
#endif
        }

        if (_PROC_ADR) {
            *CACHED_hDLL = hDLL;
#ifdef CACHE_MEMBERS
            FunctionMap[key]=(SYS_INT)_PROC_ADR;
#endif
            return((SYS_INT)_PROC_ADR);
        }
    }
    return(0);
}

void DoneLinking() {
#ifdef CACHE_MEMBERS
    FunctionMap.clear();
#endif
}
