//#define GO_DEBUG

//------------ standard header -----------------------------------//
#include "stdlibrary.h"
//------------ end of standard header ----------------------------//
#include "library.h"
#ifdef _WIN32
    #include <windows.h>
#else
    #include <pthread.h>
    #include <unistd.h>
    #define DWORD long
    #define LPVOID void*
#endif
#include "AnsiString.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "semhh.h"
#include <vector>

struct Container {
	void *DELEGATE;
	void *CONTEXT;
};

INVOKE_CALL LocalInvoker;
static int MultiThreaded = 0;
//---------------------------------------------------------------------------
CONCEPT_DLL_API ON_CREATE_CONTEXT MANAGEMENT_PARAMETERS {
    LocalInvoker=Invoke;
    MultiThreaded = Invoke(INVOKE_MULTITHREADED);
    return 0;
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API ON_DESTROY_CONTEXT MANAGEMENT_PARAMETERS {
    return 0;
}
//---------------------------------------------------------------------------
#ifdef _WIN32
DWORD WINAPI ThreadFunction(LPVOID DPARAM) {
#else
LPVOID ThreadFunction(LPVOID DPARAM) {
#endif
	Container *cdc=(Container *)DPARAM;
	void *DELEGATE=cdc->DELEGATE;
	void *context=cdc->CONTEXT;
	delete cdc;

	void *RES=0;
    void *EXCEPTION=0;
	LocalInvoker(INVOKE_SET_THREAD_DATA, context);
    LocalInvoker(INVOKE_CALL_DELEGATE_THREAD,DELEGATE,&RES,&EXCEPTION,(INTEGER)0);
    LocalInvoker(INVOKE_FREE_VARIABLE,RES);
    LocalInvoker(INVOKE_FREE_VARIABLE,EXCEPTION);
    LocalInvoker(INVOKE_FREE_VARIABLE,DELEGATE);
    return 0;
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_RunThread CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK_MIN_MAX(1, 2,"RunThread parameters error : RunThread(delegate, detached=false)");
    if (!MultiThreaded) {
        return (void *)"RunThread must be called from multi-threaded Concept core. You are now running a non-thread safe version of Concept.";
    }
    LOCAL_INIT;

    char *cls;
    NUMBER mid;
	void *context=0;
    NUMBER ndetachable=0;

    GET_CHECK_DELEGATE(0,cls,mid,"RunThread : delegate should be of DELEGATE type");
    if (PARAMETERS_COUNT>1) {
        GET_CHECK_NUMBER(1,ndetachable,"RunThread : detached should be of NUMBER type");
    }

	Container *cdc=new Container();

    //LocalInvoker(INVOKE_DYNAMIC_LOCK,PARAMETER(0));
    LocalInvoker(INVOKE_LOCK_VARIABLE,PARAMETER(0));
	LocalInvoker(INVOKE_GET_THREAD_DATA, &context);

	cdc->DELEGATE=PARAMETER(0);
	cdc->CONTEXT=context;
#ifdef _WIN32
    DWORD threadID=0;
    HANDLE thandle=CreateThread(NULL,0,ThreadFunction,(LPVOID)cdc,0,&threadID);
    if ((thandle) && ((int)ndetachable))
        CloseHandle(thandle);
    threadID=(long)thandle;
#else
    pthread_t threadID=0;
    if ((int)ndetachable) {
        pthread_attr_t attr;
        pthread_attr_init(&attr);
        pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);

        if (pthread_create(&threadID, &attr, ThreadFunction,(LPVOID)cdc))
            threadID=0;
    } else {
        if (pthread_create(&threadID, NULL, ThreadFunction,(LPVOID)cdc))
            threadID=0;
    }
#endif
    RETURN_NUMBER((double)(long)threadID);
    return 0;
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_KillThread CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK(1,"KillThread parameters error : KillThread(threadID)");

    LOCAL_INIT;
    NUMBER nThreadID;

    GET_CHECK_NUMBER(0,nThreadID,"KillThread : threadID should be of STATIC NUMBER type");

#ifdef _WIN32
    TerminateThread((HANDLE)(long)nThreadID,0);
#else
    pthread_cancel((pthread_t)(long)nThreadID);
#endif
    RETURN_NUMBER(0);
    return 0;
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_WaitThread CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK(1,"WaitThread parameters error : WaitThread(threadID)");

    LOCAL_INIT;
    NUMBER nThreadID;

    GET_CHECK_NUMBER(0,nThreadID,"WaitThread : threadID should be of STATIC NUMBER type");

#ifdef _WIN32
    WaitForSingleObject((HANDLE)(long)nThreadID,INFINITE);
    CloseHandle((HANDLE)(long)nThreadID);
#else
    void *data;
    pthread_join((pthread_t)(long)nThreadID,&data);
#endif
    RETURN_NUMBER(0);
    return 0;
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT__Sleep CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK(1,"Sleep : Sleep(miliseconds)");

    LOCAL_INIT;
    NUMBER nSeconds;

    GET_CHECK_NUMBER(0,nSeconds,"Sleep : 'miliseconds' should be of STATIC NUMBER type");

#ifdef _WIN32
    Sleep((long)nSeconds);
#else
    usleep((long)nSeconds*1000);
#endif
    RETURN_NUMBER(0);
    return 0;
}
//---------------------------------------------------------------------------
CONCEPT_FUNCTION_IMPL(_semcreate,0)
    HHSEM *sm=new HHSEM;
    RETURN_NUMBER((long)sm)
END_IMPL
//---------------------------------------------------------------------------
CONCEPT_FUNCTION_IMPL(_semdone,1)
    T_HANDLE(0)

    HHSEM *sm=(HHSEM *)PARAM_INT(0);
    semdel((*sm));
    delete sm;

    RETURN_NUMBER(0)
END_IMPL
//---------------------------------------------------------------------------
CONCEPT_FUNCTION_IMPL(_seminit,2)
    T_HANDLE(0)
    T_NUMBER(1)

    HHSEM *sm=(HHSEM *)PARAM_INT(0);

    seminit((*sm), PARAM_INT(1));

    RETURN_NUMBER(0)
END_IMPL
//---------------------------------------------------------------------------
CONCEPT_FUNCTION_IMPL(_semp,1)
    T_HANDLE(0)

    HHSEM *sm=(HHSEM *)PARAM_INT(0);

    semp((*sm));

    RETURN_NUMBER(0)
END_IMPL
//---------------------------------------------------------------------------
CONCEPT_FUNCTION_IMPL(_semv,1)
    T_HANDLE(0)

    HHSEM *sm=(HHSEM *)PARAM_INT(0);

    semv((*sm));

    RETURN_NUMBER(0)
END_IMPL
//---------------------------------------------------------------------------
CONCEPT_FUNCTION_IMPL(IsThreaded,0)
    int res=Invoke(INVOKE_MULTITHREADED);
    RETURN_NUMBER(res);
END_IMPL
//---------------------------------------------------------------------------
CONCEPT_FUNCTION_IMPL(Greenlets,1)
    T_ARRAY(0)

    std::vector<void *> *items=new std::vector<void *>;
    INTEGER count=Invoke(INVOKE_GET_ARRAY_COUNT, PARAMETER(0));
    void *prec=0;
    void *first=0;
    void *arr=PARAMETER(0);
    for (INTEGER i=0;i<count;i++) {
        void *var=0;
        Invoke(INVOKE_ARRAY_VARIABLE,arr,i,&var);
        if (var) {
            INTEGER type=0;
            char *szData=0;
            NUMBER nData=0;

            Invoke(INVOKE_GET_VARIABLE, var, &type, &szData, &nData);
            if (type==VARIABLE_DELEGATE) {
                void *ref=0;
                Invoke(INVOKE_GREENTHREAD, var, &ref);
                if (ref) {
                    Invoke(INVOKE_LOCK_VARIABLE, var);
                    items->push_back(ref);
                    if (prec)
                        Invoke(INVOKE_GREENLINK, prec, ref);
                    prec=ref;
                    if (!first)
                        first=ref;
                    Invoke(INVOKE_GREENLINK, ref, first);
                }
            }
        }
    }

    RETURN_NUMBER((SYS_INT)items);
END_IMPL
//---------------------------------------------------------------------------
CONCEPT_FUNCTION_IMPL(GreenletAdd,2)
    T_HANDLE(0)
    T_DELEGATE(1)

    std::vector<void *> *items=(std::vector<void *> *)(SYS_INT)PARAM(0);
    void *ref=0;
    int res=Invoke(INVOKE_GREENTHREAD, PARAMETER(1), &ref);
    if (ref) {
        void *first=items->front();
        // check if run-time insert is possible
        if (!IS_OK(Invoke(INVOKE_GREENINSERT, first, ref))) {
            void *prec=items->back();
            items->push_back(ref);
            if (prec)
                Invoke(INVOKE_GREENLINK, prec, ref);
            if (!first)
                first=ref;
            Invoke(INVOKE_GREENLINK, ref, first);
        }
        Invoke(INVOKE_LOCK_VARIABLE, PARAMETER(1));
    }
    RETURN_NUMBER(res);
END_IMPL
//---------------------------------------------------------------------------
CONCEPT_FUNCTION_IMPL(GreenLoop,1)
    T_HANDLE(0)

    std::vector<void *> *items=(std::vector<void *> *)(SYS_INT)PARAM(0);
    void *first=items->front();
    void *last=items->back();
    int res=-2;
    if (first)
        res=Invoke(INVOKE_GREENLOOP, first, last);

    delete items;
    SET_NUMBER(0,0);
    RETURN_NUMBER(res);
END_IMPL
//---------------------------------------------------------------------------
CONCEPT_FUNCTION_IMPL(yield,0)
    RETURN_NUMBER(0);
END_IMPL
//---------------------------------------------------------------------------
